@@ -18,11 +18,11 @@
  *)
 open Extlib_leftovers
 open Ast
-open CompilationServer
 open Type
 open Globals
 open Define
 open NativeLibraries
+open Warning
 
 type package_rule =
 	| Forbidden
@@ -46,32 +46,6 @@ type stats = {
 	s_macros_called : int ref;
 }
 
-type compiler_message =
-	| CMInfo of string * pos
-	| CMWarning of string * pos
-	| CMError of string * pos
-
-let compiler_message_string msg =
-	let (str,p) = match msg with
-		| CMInfo(str,p) | CMError(str,p) -> (str,p)
-		| CMWarning(str,p) -> ("Warning : " ^ str, p)
-	in
-	if p = null_pos then
-		str
-	else begin
-		let error_printer file line = Printf.sprintf "%s:%d:" file line in
-		let epos = Lexer.get_error_pos error_printer p in
-		let str =
-			let lines =
-				match (ExtString.String.nsplit str "\n") with
-				| first :: rest -> first :: List.map Error.compl_msg rest
-				| l -> l
-			in
-			String.concat ("\n" ^ epos ^ " : ") lines
-		in
-		Printf.sprintf "%s : %s" epos str
-	end
-
 (**
 	The capture policy tells which handling we make of captured locals
 	(the locals which are referenced in local functions)
@@ -91,6 +65,11 @@ type exceptions_config = {
 	ec_native_throws : path list;
 	(* Base types which may be caught from Haxe code without wrapping. *)
 	ec_native_catches : path list;
+	(*
+		Hint exceptions filter to avoid wrapping for targets, which can throw/catch any type
+		Ignored on targets with a specific native base type for exceptions.
+	*)
+	ec_avoid_wrapping : bool;
 	(* Path of a native class or interface, which can be used for wildcard catches. *)
 	ec_wildcard_catch : path;
 	(*
@@ -184,42 +163,53 @@ type platform_config = {
 	pf_exceptions : exceptions_config;
 	(** the scoping of local variables *)
 	pf_scoping : var_scoping_config;
+	(** target supports atomic operations via haxe.Atomic **)
+	pf_supports_atomics : bool;
 }
 
 class compiler_callbacks = object(self)
-	val mutable before_typer_create = [];
-	val mutable after_init_macros = [];
+	val before_typer_create = ref [];
+	val after_init_macros = ref [];
 	val mutable after_typing = [];
-	val mutable before_save = [];
-	val mutable after_save = [];
-	val mutable after_filters = [];
-	val mutable after_generation = [];
+	val before_save = ref [];
+	val after_save = ref [];
+	val after_filters = ref [];
+	val after_generation = ref [];
 	val mutable null_safety_report = [];
 
 	method add_before_typer_create (f : unit -> unit) : unit =
-		before_typer_create <- f :: before_typer_create
+		before_typer_create := f :: !before_typer_create
 
 	method add_after_init_macros (f : unit -> unit) : unit =
-		after_init_macros <- f :: after_init_macros
+		after_init_macros := f :: !after_init_macros
 
 	method add_after_typing (f : module_type list -> unit) : unit =
 		after_typing <- f :: after_typing
 
 	method add_before_save (f : unit -> unit) : unit =
-		before_save <- f :: before_save
+		before_save := f :: !before_save
 
 	method add_after_save (f : unit -> unit) : unit =
-		after_save <- f :: after_save
+		after_save := f :: !after_save
 
 	method add_after_filters (f : unit -> unit) : unit =
-		after_filters <- f :: after_filters
+		after_filters := f :: !after_filters
 
 	method add_after_generation (f : unit -> unit) : unit =
-		after_generation <- f :: after_generation
+		after_generation := f :: !after_generation
 
 	method add_null_safety_report (f : (string*pos) list -> unit) : unit =
 		null_safety_report <- f :: null_safety_report
 
+	method run r =
+		match !r with
+		| [] ->
+			()
+		| l ->
+			r := [];
+			List.iter (fun f -> f()) (List.rev l);
+			self#run r
+
 	method get_before_typer_create = before_typer_create
 	method get_after_init_macros = after_init_macros
 	method get_after_typing = after_typing
@@ -243,7 +233,7 @@ class file_keys = object(self)
 end
 
 type shared_display_information = {
-	mutable diagnostics_messages : (string * pos * DisplayTypes.DiagnosticsKind.t * DisplayTypes.DiagnosticsSeverity.t) list;
+	mutable diagnostics_messages : (string * pos * MessageKind.t * MessageSeverity.t * int (* depth *)) list;
 }
 
 type display_information = {
@@ -281,14 +271,83 @@ type compiler_stage =
 	| CGenerationStart  (* Generation is about to begin. *)
 	| CGenerationDone   (* Generation just finished. *)
 
+type report_mode =
+	| RMNone
+	| RMDiagnostics of Path.UniqueKey.t list
+	| RMStatistics
+
+class virtual ['key,'value] lookup = object(self)
+	method virtual add : 'key -> 'value -> unit
+	method virtual remove : 'key -> unit
+	method virtual find : 'key -> 'value
+	method virtual iter : ('key -> 'value -> unit) -> unit
+	method virtual fold : 'acc . ('key -> 'value -> 'acc -> 'acc) -> 'acc -> 'acc
+	method virtual mem : 'key -> bool
+	method virtual clear : unit
+end
+
+class ['key,'value] pmap_lookup = object(self)
+	inherit ['key,'value] lookup
+	val mutable lut : ('key,'value) PMap.t = PMap.empty
+
+	method add (key : 'key) (value : 'value) =
+		lut <- PMap.add key value lut
+
+	method remove (key : 'key) =
+		lut <- PMap.remove key lut
+
+	method find (key : 'key) : 'value =
+		PMap.find key lut
+
+	method iter (f : 'key -> 'value -> unit) =
+		PMap.iter f lut
+
+	method fold : 'acc . ('key -> 'value -> 'acc -> 'acc) -> 'acc -> 'acc = fun f acc ->
+		PMap.foldi f lut acc
+
+	method mem (key : 'key) =
+		PMap.mem key lut
+
+	method clear =
+		lut <- PMap.empty
+end
+
+class ['key,'value] hashtbl_lookup = object(self)
+	inherit ['key,'value] lookup
+	val lut : ('key,'value) Hashtbl.t = Hashtbl.create 0
+
+	method add (key : 'key) (value : 'value) =
+		Hashtbl.replace lut key value
+
+	method remove (key : 'key) =
+		Hashtbl.remove lut key
+
+	method find (key : 'key) : 'value =
+		Hashtbl.find lut key
+
+	method iter (f : 'key -> 'value -> unit) =
+		Hashtbl.iter f lut
+
+	method fold : 'acc . ('key -> 'value -> 'acc -> 'acc) -> 'acc -> 'acc = fun f acc ->
+		Hashtbl.fold f lut acc
+
+	method mem (key : 'key) =
+		Hashtbl.mem lut key
+
+	method clear =
+		Hashtbl.clear lut
+end
+
 type context = {
+	compilation_step : int;
 	mutable stage : compiler_stage;
-	mutable cache : context_cache option;
+	cs : CompilationCache.t;
+	mutable cache : CompilationCache.context_cache option;
+	is_macro_context : bool;
+	mutable json_out : json_api option;
 	(* config *)
 	version : int;
 	args : string list;
-	shared : shared_context;
-	display_information : display_information;
 	mutable sys_args : string list;
 	mutable display : DisplayTypes.DisplayMode.settings;
 	mutable debug : bool;
@@ -300,48 +359,69 @@ type context = {
 	mutable class_path : string list;
 	mutable main_class : path option;
 	mutable package_rules : (string,package_rule) PMap.t;
-	mutable error : string -> pos -> unit;
-	mutable info : string -> pos -> unit;
-	mutable warning : string -> pos -> unit;
+	mutable report_mode : report_mode;
+	(* communication *)
+	mutable print : string -> unit;
+	mutable error : ?depth:int -> string -> pos -> unit;
+	mutable located_error : ?depth:int -> located -> unit;
+	mutable info : ?depth:int -> string -> pos -> unit;
+	mutable warning : ?depth:int -> warning -> Warning.warning_option list list -> string -> pos -> unit;
+	mutable warning_options : Warning.warning_option list list;
 	mutable get_messages : unit -> compiler_message list;
 	mutable filter_messages : (compiler_message -> bool) -> unit;
+	mutable run_command : string -> int;
+	mutable run_command_args : string -> string list -> int;
+	(* typing setup *)
 	mutable load_extern_type : (string * (path -> pos -> Ast.package option)) list; (* allow finding types which are not in sources *)
 	callbacks : compiler_callbacks;
 	defines : Define.define;
-	mutable print : string -> unit;
+	mutable user_defines : (string, Define.user_define) Hashtbl.t;
+	mutable user_metas : (string, Meta.user_meta) Hashtbl.t;
 	mutable get_macros : unit -> context option;
-	mutable run_command : string -> int;
-	file_lookup_cache : (string,string option) Hashtbl.t;
+	(* typing state *)
+	shared : shared_context;
+	display_information : display_information;
+	file_lookup_cache : (string,string option) lookup;
 	file_keys : file_keys;
-	readdir_cache : (string * string,(string array) option) Hashtbl.t;
-	parser_cache : (string,(type_def * pos) list) Hashtbl.t;
-	module_to_file : (path,string) Hashtbl.t;
-	cached_macros : (path * string,(((string * bool * t) list * t * tclass * Type.tclass_field) * module_def)) Hashtbl.t;
-	mutable stored_typed_exprs : (int, texpr) PMap.t;
+	readdir_cache : (string * string,(string array) option) lookup;
+	parser_cache : (string,(type_def * pos) list) lookup;
+	module_to_file : (path,string) lookup;
+	cached_macros : (path * string,(((string * bool * t) list * t * tclass * Type.tclass_field) * module_def)) lookup;
+	stored_typed_exprs : (int, texpr) lookup;
+	overload_cache : ((path * string),(Type.t * tclass_field) list) lookup;
+	module_lut : (path,module_def) lookup;
+	module_nonexistent_lut : (path,bool) lookup;
+	type_to_module : (path,path) lookup;
+	mutable has_error : bool;
 	pass_debug_messages : string DynArray.t;
-	overload_cache : ((path * string),(Type.t * tclass_field) list) Hashtbl.t;
 	(* output *)
 	mutable file : string;
-	mutable flash_version : float;
 	mutable features : (string,bool) Hashtbl.t;
 	mutable modules : Type.module_def list;
 	mutable main : Type.texpr option;
 	mutable types : Type.module_type list;
 	mutable resources : (string,string) Hashtbl.t;
-	mutable neko_libs : string list;
+	(* target-specific *)
+	mutable flash_version : float;
+	mutable neko_lib_paths : string list;
 	mutable include_files : (string * string) list;
 	mutable native_libs : native_libraries;
 	mutable net_std : string list;
 	net_path_map : (path,string list * string list * string) Hashtbl.t;
 	mutable c_args : string list;
 	mutable js_gen : (unit -> unit) option;
-	mutable json_out : json_api option;
-	(* typing *)
+	(* misc *)
 	mutable basic : basic_types;
 	memory_marker : float array;
 }
 
-exception Abort of string * pos
+exception Abort of located
+
+let ignore_error com =
+	let b = com.display.dms_error_policy = EPIgnore in
+	if b then
+		if b then com.has_error <- true;
+	b
 
 (* Defines *)
 
@@ -370,8 +450,58 @@ let raw_define com v =
 let define_value com k v =
 	Define.define_value com.defines k v
 
-let raw_defined_value com k =
-	Define.raw_defined_value com.defines k
+let convert_define k =
+	String.concat "_" (ExtString.String.nsplit k "-")
+
+let is_next com = defined com HaxeNext
+
+let external_defined ctx k =
+	Define.raw_defined ctx.defines (convert_define k)
+
+let external_defined_value ctx k =
+	Define.raw_defined_value ctx.defines (convert_define k)
+
+let reserved_flags = [
+	"true";"false";"null";"cross";"js";"lua";"neko";"flash";"php";"cpp";"cs";"java";"python";
+	"swc";"macro";"sys";"static";"utf16";"haxe";"haxe_ver"
+	]
+
+let reserved_flag_namespaces = ["target"]
+
+let convert_and_validate k =
+	let converted_flag = convert_define k in
+	let raise_reserved description =
+		raise (Arg.Bad (description ^ " and cannot be defined from the command line"))
+	in
+	if List.mem converted_flag reserved_flags then
+		raise_reserved (Printf.sprintf "`%s` is a reserved compiler flag" k);
+	List.iter (fun ns ->
+		if ExtString.String.starts_with converted_flag (ns ^ ".") then
+			raise_reserved (Printf.sprintf "`%s` uses the reserved compiler flag namespace `%s.*`" k ns)
+	) reserved_flag_namespaces;
+	converted_flag
+
+let external_define_value ctx k v =
+	raw_define_value ctx.defines (convert_and_validate k) v
+
+(* TODO: Temporary function until #8690, remove after *)
+let external_define_value_no_check ctx k v =
+	Define.raw_define_value ctx.defines (convert_define k) v
+
+let external_define ctx k =
+	Define.raw_define ctx.defines (convert_and_validate k)
+
+(* TODO: Temporary function until #8690, remove after *)
+let external_define_no_check ctx k =
+	Define.raw_define ctx.defines (convert_define k)
+
+let defines_for_external ctx =
+	PMap.foldi (fun k v acc ->
+		let added_underscore = PMap.add k v acc in
+		match ExtString.String.nsplit k "_" with
+			| [_] -> added_underscore
+			| split -> PMap.add (String.concat "-" split) v added_underscore;
+	) ctx.defines.values PMap.empty
 
 let get_es_version com =
 	try int_of_string (defined_value com Define.JsEs) with _ -> 0
@@ -419,16 +549,18 @@ let default_config =
 			ec_native_catches = [];
 			ec_wildcard_catch = (["StdTypes"],"Dynamic");
 			ec_base_throw = (["StdTypes"],"Dynamic");
+			ec_avoid_wrapping = true;
 			ec_special_throw = fun _ -> false;
 		};
 		pf_scoping = {
 			vs_scope = BlockScope;
 			vs_flags = [];
-		}
+		};
+		pf_supports_atomics = false;
 	}
 
 let get_config com =
-	let defined f = PMap.mem (fst (Define.infos f)) com.defines.values in
+	let defined f = PMap.mem (Define.get_define_key f) com.defines.values in
 	match com.platform with
 	| Cross ->
 		default_config
@@ -447,13 +579,15 @@ let get_config com =
 					["js";"lib"],"Error";
 					["haxe"],"Exception";
 				];
+				ec_avoid_wrapping = false;
 			};
 			pf_scoping = {
 				vs_scope = if es6 then BlockScope else FunctionScope;
 				vs_flags =
 					(if defined Define.JsUnflatten then ReserveAllTopLevelSymbols else ReserveAllTypesFlat)
 					:: if es6 then [NoShadowing; SwitchCasesNoBlocks;] else [VarHoisting; NoCatchVarShadowing];
-			}
+			};
+			pf_supports_atomics = true;
 		}
 	| Lua ->
 		{
@@ -462,6 +596,9 @@ let get_config com =
 			pf_capture_policy = CPLoopVars;
 			pf_uses_utf16 = false;
 			pf_supports_rest_args = true;
+			pf_exceptions = { default_config.pf_exceptions with
+				ec_avoid_wrapping = false;
+			}
 		}
 	| Neko ->
 		{
@@ -492,6 +629,7 @@ let get_config com =
 					["flash";"errors"],"Error";
 					["haxe"],"Exception";
 				];
+				ec_avoid_wrapping = false;
 			};
 			pf_scoping = {
 				vs_scope = FunctionScope;
@@ -530,8 +668,10 @@ let get_config com =
 			pf_supports_threads = true;
 			pf_supports_unicode = (defined Define.Cppia) || not (defined Define.DisableUnicodeStrings);
 			pf_scoping = { default_config.pf_scoping with
-				vs_flags = [NoShadowing]
-			}
+				vs_flags = [NoShadowing];
+				vs_scope = FunctionScope;
+			};
+			pf_supports_atomics = true;
 		}
 	| Cs ->
 		{
@@ -542,7 +682,7 @@ let get_config com =
 			pf_supports_threads = true;
 			pf_supports_rest_args = true;
 			pf_this_before_super = false;
-			pf_exceptions = {
+			pf_exceptions = { default_config.pf_exceptions with
 				ec_native_throws = [
 					["cs";"system"],"Exception";
 					["haxe"],"Exception";
@@ -559,6 +699,7 @@ let get_config com =
 				vs_scope = FunctionScope;
 				vs_flags = [NoShadowing]
 			};
+			pf_supports_atomics = true;
 		}
 	| Java ->
 		{
@@ -588,7 +729,8 @@ let get_config com =
 					{
 						vs_scope = FunctionScope;
 						vs_flags = [NoShadowing; ReserveAllTopLevelSymbols; ReserveNames(["_"])];
-					}
+					};
+			pf_supports_atomics = true;
 		}
 	| Python ->
 		{
@@ -619,14 +761,7 @@ let get_config com =
 			pf_capture_policy = CPWrapRef;
 			pf_pad_nulls = true;
 			pf_supports_threads = true;
-			pf_exceptions = { default_config.pf_exceptions with
-				ec_native_throws = [
-					["haxe"],"Exception";
-				];
-				ec_native_catches = [
-					["haxe"],"Exception";
-				];
-			}
+			pf_supports_atomics = true;
 		}
 	| Eval ->
 		{
@@ -640,14 +775,11 @@ let get_config com =
 
 let memory_marker = [|Unix.time()|]
 
-let create version s_version args =
+let create compilation_step cs version args =
 	let m = Type.mk_mono() in
-	let defines =
-		PMap.add "true" "1" (
-			PMap.add "source_header" ("Generated by Haxe " ^ s_version) PMap.empty
-		)
-	in
-	{
+	let rec com = {
+		compilation_step = compilation_step;
+		cs = cs;
 		cache = None;
 		stage = CCreated;
 		version = version;
@@ -672,6 +804,7 @@ let create version s_version args =
 		config = default_config;
 		print = (fun s -> print_string s; flush stdout);
 		run_command = Sys.command;
+		run_command_args = (fun s args -> com.run_command (Printf.sprintf "%s %s" s (String.concat " " args)));
 		std_path = [];
 		class_path = [];
 		main_class = None;
@@ -680,6 +813,9 @@ let create version s_version args =
 		types = [];
 		callbacks = new compiler_callbacks;
 		modules = [];
+		module_lut = new hashtbl_lookup;
+		module_nonexistent_lut = new hashtbl_lookup;
+		type_to_module = new hashtbl_lookup;
 		main = None;
 		flash_version = 10.;
 		resources = Hashtbl.create 0;
@@ -687,56 +823,69 @@ let create version s_version args =
 		native_libs = create_native_libs();
 		net_path_map = Hashtbl.create 0;
 		c_args = [];
-		neko_libs = [];
+		neko_lib_paths = [];
 		include_files = [];
 		js_gen = None;
 		load_extern_type = [];
 		defines = {
 			defines_signature = None;
-			values = defines;
+			values = PMap.empty;
 		};
+		user_defines = Hashtbl.create 0;
+		user_metas = Hashtbl.create 0;
 		get_macros = (fun() -> None);
-		info = (fun _ _ -> die "" __LOC__);
-		warning = (fun _ _ -> die "" __LOC__);
-		error = (fun _ _ -> die "" __LOC__);
+		info = (fun ?depth _ _ -> die "" __LOC__);
+		warning = (fun ?depth _ _ _ -> die "" __LOC__);
+		warning_options = [];
+		error = (fun ?depth _ _ -> die "" __LOC__);
+		located_error = (fun ?depth _ -> die "" __LOC__);
 		get_messages = (fun() -> []);
 		filter_messages = (fun _ -> ());
 		pass_debug_messages = DynArray.create();
-		overload_cache = Hashtbl.create 0;
 		basic = {
 			tvoid = m;
 			tint = m;
 			tfloat = m;
 			tbool = m;
-			tnull = (fun _ -> die "" __LOC__);
+			tnull = (fun _ -> die "Could use locate abstract Null<T> (was it redefined?)" __LOC__);
 			tstring = m;
-			tarray = (fun _ -> die "" __LOC__);
+			tarray = (fun _ -> die "Could not locate class Array<T> (was it redefined?)" __LOC__);
 		};
-		file_lookup_cache = Hashtbl.create 0;
+		file_lookup_cache = new hashtbl_lookup;
 		file_keys = new file_keys;
-		readdir_cache = Hashtbl.create 0;
-		module_to_file = Hashtbl.create 0;
-		stored_typed_exprs = PMap.empty;
-		cached_macros = Hashtbl.create 0;
+		readdir_cache = new hashtbl_lookup;
+		module_to_file = new hashtbl_lookup;
+		stored_typed_exprs = new hashtbl_lookup;
+		cached_macros = new hashtbl_lookup;
 		memory_marker = memory_marker;
-		parser_cache = Hashtbl.create 0;
+		parser_cache = new hashtbl_lookup;
+		overload_cache = new hashtbl_lookup;
 		json_out = None;
-	}
+		has_error = false;
+		report_mode = RMNone;
+		is_macro_context = false;
+	} in
+	com
+
+let is_diagnostics com = match com.report_mode with
+	| RMDiagnostics _ -> true
+	| _ -> false
+
+let disable_report_mode com =
+	let old = com.report_mode in
+	com.report_mode <- RMNone;
+	(fun () -> com.report_mode <- old)
 
 let log com str =
 	if com.verbose then com.print (str ^ "\n")
 
-let clone com =
+let clone com is_macro_context =
 	let t = com.basic in
 	{ com with
 		cache = None;
 		basic = { t with tvoid = t.tvoid };
 		main_class = None;
 		features = Hashtbl.create 0;
-		file_lookup_cache = Hashtbl.create 0;
-		readdir_cache = Hashtbl.create 0;
-		parser_cache = Hashtbl.create 0;
-		module_to_file = Hashtbl.create 0;
 		callbacks = new compiler_callbacks;
 		display_information = {
 			unresolved_identifiers = [];
@@ -748,16 +897,18 @@ let clone com =
 			defines_signature = com.defines.defines_signature;
 		};
 		native_libs = create_native_libs();
-		overload_cache = Hashtbl.create 0;
+		is_macro_context = is_macro_context;
+		file_lookup_cache = new hashtbl_lookup;
+		readdir_cache = new hashtbl_lookup;
+		parser_cache = new hashtbl_lookup;
+		module_to_file = new hashtbl_lookup;
+		overload_cache = new hashtbl_lookup;
+		module_lut = new hashtbl_lookup;
+		type_to_module = new hashtbl_lookup;
 	}
 
 let file_time file = Extc.filetime file
 
-let file_extension file =
-	match List.rev (ExtString.String.nsplit file ".") with
-	| e :: _ -> String.lowercase e
-	| [] -> ""
-
 let flash_versions = List.map (fun v ->
 	let maj = int_of_float v in
 	let min = int_of_float (mod_float (v *. 10.) 10.) in
@@ -792,26 +943,38 @@ let init_platform com pf =
 	com.package_rules <- List.fold_left forbid com.package_rules ("jvm" :: (List.map platform_name platforms));
 	com.config <- get_config com;
 	if com.config.pf_static then begin
-		raw_define_value com.defines "target.static" "true";
+		raw_define com "target.static";
 		define com Define.Static;
 	end;
 	if com.config.pf_sys then begin
-		raw_define_value com.defines "target.sys" "true";
+		raw_define com "target.sys";
 		define com Define.Sys
 	end else
 		com.package_rules <- PMap.add "sys" Forbidden com.package_rules;
 	if com.config.pf_uses_utf16 then begin
-		raw_define_value com.defines "target.utf16" "true";
+		raw_define com "target.utf16";
 		define com Define.Utf16;
 	end;
 	if com.config.pf_supports_threads then begin
-		raw_define_value com.defines "target.threaded" "true";
+		raw_define com "target.threaded";
 	end;
 	if com.config.pf_supports_unicode then begin
-		raw_define_value com.defines "target.unicode" "true";
+		raw_define com "target.unicode";
 	end;
 	raw_define_value com.defines "target.name" name;
-	raw_define com name
+	raw_define com name;
+	if com.config.pf_supports_atomics then begin
+		raw_define com "target.atomics"
+	end
+
+let set_platform com pf file =
+	if com.platform <> Cross then failwith "Multiple targets";
+	init_platform com pf;
+	com.file <- file;
+	if (pf = Flash) && Path.file_extension file = "swc" then define com Define.Swc;
+	(* Set the source header, unless the user has set one already or the platform sets a custom one *)
+	if not (defined com Define.SourceHeader) && (pf <> Hl) then
+		define_value com Define.SourceHeader ("Generated by Haxe " ^ s_version_full)
 
 let add_feature com f =
 	Hashtbl.replace com.features f true
@@ -863,31 +1026,32 @@ let allow_package ctx s =
 	with Not_found ->
 		()
 
-let abort msg p = raise (Abort (msg,p))
+let abort_located ?depth msg = raise (Abort msg)
+let abort ?depth msg p = abort_located ~depth (located msg p)
 
 let platform ctx p = ctx.platform = p
 
 let platform_name_macro com =
 	if defined com Define.Macro then "macro" else platform_name com.platform
 
+let remove_extension file =
+	try String.sub file 0 (String.rindex file '.')
+	with Not_found -> file
+
+let extension file =
+	try
+		let dot_pos = String.rindex file '.' in
+		String.sub file dot_pos (String.length file - dot_pos)
+	with Not_found -> file
+
 let cache_directory ctx class_path dir f_dir =
 	let platform_ext = "." ^ (platform_name_macro ctx)
 	and is_loading_core_api = defined ctx Define.CoreApi in
-	let remove_extension file =
-		try String.sub file 0 (String.rindex file '.')
-		with Not_found -> file
-	in
-	let extension file =
-		try
-			let dot_pos = String.rindex file '.' in
-			String.sub file dot_pos (String.length file - dot_pos)
-		with Not_found -> file
-	in
 	let dir_listing =
 		try Some (Sys.readdir dir);
 		with Sys_error _ -> None
 	in
-	Hashtbl.add ctx.readdir_cache (class_path,dir) dir_listing;
+	ctx.readdir_cache#add (class_path,dir) dir_listing;
 	(*
 		This function is invoked for each file in the `dir`.
 		Each file is checked if it's specific for current platform
@@ -921,23 +1085,23 @@ let cache_directory ctx class_path dir f_dir =
 			- or this is a platform-specific file for `representation`
 			- this `representation` was never found before
 		*)
-		if is_loading_core_api || is_platform_specific || not (Hashtbl.mem ctx.file_lookup_cache representation) then begin
+		if is_loading_core_api || is_platform_specific || not (ctx.file_lookup_cache#mem representation) then begin
 			let full_path = if dir = "." then file_own_name else dir ^ "/" ^ file_own_name in
-			Hashtbl.replace ctx.file_lookup_cache representation (Some full_path);
+			ctx.file_lookup_cache#add representation (Some full_path);
 		end
 	in
 	Option.may (Array.iter prepare_file) dir_listing
 
 let find_file ctx f =
 	try
-		match Hashtbl.find ctx.file_lookup_cache f with
+		match ctx.file_lookup_cache#find f with
 		| None -> raise Exit
 		| Some f -> f
 	with
 	| Exit ->
 		raise Not_found
 	| Not_found when Path.is_absolute_path f ->
-		Hashtbl.add ctx.file_lookup_cache f (Some f);
+		ctx.file_lookup_cache#add f (Some f);
 		f
 	| Not_found ->
 		let f_dir = Filename.dirname f in
@@ -949,13 +1113,13 @@ let find_file ctx f =
 				let dir = Filename.dirname file in
 				(* If we have seen the directory before, we can assume that the file isn't in there because the else case
 				   below would have added it to `file_lookup_cache`, which we check before we get here. *)
-				if Hashtbl.mem ctx.readdir_cache (p,dir) then
+				if ctx.readdir_cache#mem (p,dir) then
 					loop (had_empty || p = "") l
 				else begin
 					cache_directory ctx p dir f_dir;
 					(* Caching might have located the file we're looking for, so check the lookup cache again. *)
 					try
-						begin match Hashtbl.find ctx.file_lookup_cache f with
+						begin match ctx.file_lookup_cache#find f with
 						| Some f -> f
 						| None -> raise Not_found
 						end
@@ -964,7 +1128,7 @@ let find_file ctx f =
 				end
 		in
 		let r = try Some (loop false ctx.class_path) with Not_found -> None in
-		Hashtbl.add ctx.file_lookup_cache f r;
+		ctx.file_lookup_cache#add f r;
 		match r with
 		| None -> raise Not_found
 		| Some f -> f
@@ -1067,35 +1231,48 @@ let utf16_to_utf8 str =
 	loop 0;
 	Buffer.contents b
 
-let add_diagnostics_message com s p kind sev =
+let add_diagnostics_message ?(depth = 0) com msg kind sev =
+	if sev = MessageSeverity.Error then com.has_error <- true;
 	let di = com.shared.shared_display_information in
-	di.diagnostics_messages <- (s,p,kind,sev) :: di.diagnostics_messages
+	match (extract_located msg) with
+	| [] -> ()
+	| (s,p) :: [] ->
+		di.diagnostics_messages <- (s,p,kind,sev,depth) :: di.diagnostics_messages
+	| (s,p) :: stack ->
+		let stack_diag = (List.map (fun (s,p) -> (s,p,kind,sev,depth+1)) (List.rev stack)) in
+		di.diagnostics_messages <- stack_diag @ ((s,p,kind,sev,depth) :: di.diagnostics_messages)
+
+let located_display_error com ?(depth = 0) msg =
+	if is_diagnostics com then
+		add_diagnostics_message ~depth com msg MessageKind.DKCompilerMessage MessageSeverity.Error
+	else
+		com.located_error msg ~depth
+
+let display_error com ?(depth = 0) msg p =
+	located_display_error com ~depth (Globals.located msg p)
 
 open Printer
 
-let dump_context com = s_record_fields "" [
-	"version",string_of_int com.version;
-	"args",s_list ", " (fun s -> s) com.args;
-	"debug",string_of_bool com.debug;
-	"platform",platform_name com.platform;
-	"std_path",s_list ", " (fun s -> s) com.std_path;
-	"class_path",s_list ", " (fun s -> s) com.class_path;
-	"defines",s_pmap (fun s -> s) (fun s -> s) com.defines.values;
-	"defines_signature",s_opt (fun s -> Digest.to_hex s) com.defines.defines_signature;
-]
-
 let dump_path com =
 	Define.defined_value_safe ~default:"dump" com.defines Define.DumpPath
 
 let adapt_defines_to_macro_context defines =
-	let values = ref defines.values in
-	List.iter (fun p -> values := PMap.remove (Globals.platform_name p) !values) Globals.platforms;
-	let to_remove = List.map (fun d -> fst (Define.infos d)) [Define.NoTraces] in
-	let to_remove = to_remove @ List.map (fun (_,d) -> "flash" ^ d) flash_versions in
-	values := PMap.foldi (fun k v acc -> if List.mem k to_remove then acc else PMap.add k v acc) !values PMap.empty;
-	values := PMap.add "macro" "1" !values;
-	values := PMap.add (platform_name !Globals.macro_platform) "1" !values;
-	{values = !values; defines_signature = None }
+	let to_remove = List.map Globals.platform_name Globals.platforms in
+	let to_remove = List.fold_left (fun acc d -> Define.get_define_key d :: acc) to_remove [Define.NoTraces] in
+	let to_remove = List.fold_left (fun acc (_, d) -> ("flash" ^ d) :: acc) to_remove flash_versions in
+	let macro_defines = {
+		values = PMap.foldi (fun k v acc ->
+			if List.mem k to_remove then acc else PMap.add k v acc) defines.values PMap.empty;
+		defines_signature = None
+	} in
+	Define.define macro_defines Define.Macro;
+	Define.raw_define macro_defines (platform_name !Globals.macro_platform);
+	macro_defines
+
+let adapt_defines_to_display_context defines =
+	let defines = adapt_defines_to_macro_context defines in
+	Define.define defines Define.Display;
+	defines
 
 let is_legacy_completion com = match com.json_out with
 	| None -> true
