@@ -66,10 +66,69 @@ let rec add_final_return e =
 			{ e with eexpr = TFunction f }
 		| _ -> e
 
+module LocalStatic = struct
+	let promote_local_static ctx lut v eo =
+		let name = Printf.sprintf "%s_%s" ctx.curfield.cf_name v.v_name in
+		begin try
+			let cf = PMap.find name ctx.curclass.cl_statics in
+			display_error ctx.com (Printf.sprintf "The expanded name of this local (%s) conflicts with another static field" name) v.v_pos;
+			typing_error ~depth:1 "Conflicting field was found here" cf.cf_name_pos;
+		with Not_found ->
+			let cf = mk_field name ~static:true v.v_type v.v_pos v.v_pos in
+			begin match eo with
+			| None ->
+				()
+			| Some e ->
+				let rec loop e = match e.eexpr with
+					| TLocal _ | TFunction _ ->
+						typing_error "Accessing local variables in static initialization is not allowed" e.epos
+					| TConst (TThis | TSuper) ->
+						typing_error "Accessing `this` in static initialization is not allowed" e.epos
+					| TReturn _ | TBreak | TContinue ->
+						typing_error "This kind of control flow in static initialization is not allowed" e.epos
+					| _ ->
+						iter loop e
+				in
+				loop e;
+				cf.cf_expr <- Some e
+			end;
+			TClass.add_field ctx.curclass cf;
+			Hashtbl.add lut v.v_id cf
+		end
+
+	let find_local_static lut v =
+		Hashtbl.find lut v.v_id
+
+	let run ctx e =
+		let local_static_lut = Hashtbl.create 0 in
+		let c = ctx.curclass in
+		let rec run e = match e.eexpr with
+			| TBlock el ->
+				let el = ExtList.List.filter_map (fun e -> match e.eexpr with
+					| TVar(v,eo) when has_var_flag v VStatic ->
+						promote_local_static ctx local_static_lut v eo;
+						None
+					| _ ->
+						Some (run e)
+				) el in
+				{ e with eexpr = TBlock el }
+			| TLocal v when has_var_flag v VStatic ->
+				begin try
+					let cf = find_local_static local_static_lut v in
+					Texpr.Builder.make_static_field c cf e.epos
+				with Not_found ->
+					typing_error (Printf.sprintf "Could not find local static %s (id %i)" v.v_name v.v_id) e.epos
+				end
+			| _ ->
+				Type.map_expr run e
+		in
+		run e
+end
+
 (* -------------------------------------------------------------------------- *)
 (* CHECK LOCAL VARS INIT *)
 
-let check_local_vars_init com e =
+let check_local_vars_init ctx e =
 	let intersect vl1 vl2 =
 		PMap.mapi (fun v t -> t && PMap.find v vl2) vl1
 	in
@@ -84,17 +143,21 @@ let check_local_vars_init com e =
 	in
 	let declared = ref [] in
 	let outside_vars = ref IntMap.empty in
+	(* Set variables which belong to current function *)
+	let set_all_vars vars =
+		vars := PMap.mapi (fun id is_set -> if IntMap.mem id !outside_vars then is_set else true) !vars
+	in
 	let rec loop vars e =
 		match e.eexpr with
 		| TLocal v ->
 			let init = (try PMap.find v.v_id !vars with Not_found -> true) in
 			if not init then begin
 				if IntMap.mem v.v_id !outside_vars then
-					if v.v_name = "this" then com.warning "this might be used before assigning a value to it" e.epos
-					else com.warning ("Local variable " ^ v.v_name ^ " might be used before being initialized") e.epos
+					if v.v_name = "this" then warning ctx WVarInit "this might be used before assigning a value to it" e.epos
+					else warning ctx WVarInit ("Local variable " ^ v.v_name ^ " might be used before being initialized") e.epos
 				else
-					if v.v_name = "this" then error "Missing this = value" e.epos
-					else error ("Local variable " ^ v.v_name ^ " used without being initialized") e.epos
+					if v.v_name = "this" then typing_error "Missing this = value" e.epos
+					else typing_error ("Local variable " ^ v.v_name ^ " used without being initialized") e.epos
 			end
 		| TVar (v,eo) ->
 			begin
@@ -115,8 +178,16 @@ let check_local_vars_init com e =
 			restore vars old_vars (List.rev !declared);
 			declared := old;
 		| TBinop (OpAssign,{ eexpr = TLocal v },e) when PMap.mem v.v_id !vars ->
-			loop vars e;
-			vars := PMap.add v.v_id true !vars
+			begin match (Texpr.skip e).eexpr with
+				| TFunction _ ->
+					(* We can be sure that the function doesn't execute immediately, so it's fine to
+					   consider the local initialized (issue #9919). *)
+					vars := PMap.add v.v_id true !vars;
+					loop vars e;
+				| _ ->
+					loop vars e;
+					vars := PMap.add v.v_id true !vars
+			end
 		| TIf (e1,e2,eo) ->
 			loop vars e1;
 			let vbase = !vars in
@@ -174,10 +245,10 @@ let check_local_vars_init com e =
 				join vars cvars)
 		(* mark all reachable vars as initialized, since we don't exit the block  *)
 		| TBreak | TContinue | TReturn None ->
-			vars := PMap.map (fun _ -> true) !vars
+			set_all_vars vars
 		| TThrow e | TReturn (Some e) ->
 			loop vars e;
-			vars := PMap.map (fun _ -> true) !vars
+			set_all_vars vars
 		| TFunction tf ->
 			let old = !outside_vars in
 			(* Mark all known variables as "outside" so we can ignore their initialization state within the function.
@@ -194,7 +265,7 @@ let check_local_vars_init com e =
 
 let mark_switch_break_loops e =
 	let add_loop_label n e =
-		{ e with eexpr = TMeta ((Meta.LoopLabel,[(EConst(Int(string_of_int n)),e.epos)],e.epos), e) }
+		{ e with eexpr = TMeta ((Meta.LoopLabel,[(EConst(Int(string_of_int n, None)),e.epos)],e.epos), e) }
 	in
 	let in_switch = ref false in
 	let did_found = ref (-1) in
@@ -274,7 +345,7 @@ let check_abstract_as_value e =
 		match e.eexpr with
 		| TField ({ eexpr = TTypeExpr _ }, _) -> ()
 		| TTypeExpr(TClassDecl {cl_kind = KAbstractImpl a}) when not (Meta.has Meta.RuntimeValue a.a_meta) ->
-			error "Cannot use abstract as value" e.epos
+			typing_error "Cannot use abstract as value" e.epos
 		| _ -> Type.iter loop e
 	in
 	loop e;
@@ -282,70 +353,9 @@ let check_abstract_as_value e =
 
 (* PASS 1 end *)
 
-(* Saves a class state so it can be restored later, e.g. after DCE or native path rewrite *)
-let save_class_state ctx t = match t with
-	| TClassDecl c ->
-		let vars = ref [] in
-		let rec save_vars e =
-			let add v = vars := (v, v.v_type) :: !vars in
-			match e.eexpr with
-				| TFunction fn ->
-					List.iter (fun (v, _) -> add v) fn.tf_args;
-					save_vars fn.tf_expr
-				| TVar (v, e) ->
-					add v;
-					Option.may save_vars e
-				| _ ->
-					iter save_vars e
-		in
-		let mk_field_restore f =
-			Option.may save_vars f.cf_expr;
-			let rec mk_overload_restore f =
-				f.cf_name,f.cf_kind,f.cf_expr,f.cf_type,f.cf_meta,f.cf_params
-			in
-			( f,mk_overload_restore f, List.map (fun f -> f,mk_overload_restore f) f.cf_overloads )
-		in
-		let restore_field (f,res,overloads) =
-			let restore_field (f,(name,kind,expr,t,meta,params)) =
-				f.cf_name <- name; f.cf_kind <- kind; f.cf_expr <- expr; f.cf_type <- t; f.cf_meta <- meta; f.cf_params <- params;
-				f
-			in
-			let f = restore_field (f,res) in
-			f.cf_overloads <- List.map restore_field overloads;
-			f
-		in
-		let mk_pmap lst =
-			List.fold_left (fun pmap f -> PMap.add f.cf_name f pmap) PMap.empty lst
-		in
-
-		let meta = c.cl_meta and path = c.cl_path and ext = (has_class_flag c CExtern) in
-		let sup = c.cl_super and impl = c.cl_implements in
-		let csr = Option.map (mk_field_restore) c.cl_constructor in
-		let ofr = List.map (mk_field_restore) c.cl_ordered_fields in
-		let osr = List.map (mk_field_restore) c.cl_ordered_statics in
-		let init = c.cl_init in
-		Option.may save_vars init;
-		c.cl_restore <- (fun() ->
-			c.cl_super <- sup;
-			c.cl_implements <- impl;
-			c.cl_meta <- meta;
-			if ext then add_class_flag c CExtern else remove_class_flag c CExtern;
-			c.cl_path <- path;
-			c.cl_init <- init;
-			c.cl_ordered_fields <- List.map restore_field ofr;
-			c.cl_ordered_statics <- List.map restore_field osr;
-			c.cl_fields <- mk_pmap c.cl_ordered_fields;
-			c.cl_statics <- mk_pmap c.cl_ordered_statics;
-			c.cl_constructor <- Option.map restore_field csr;
-			c.cl_descendants <- [];
-			List.iter (fun (v, t) -> v.v_type <- t) !vars;
-		)
-	| _ ->
-		()
-
 (* PASS 2 begin *)
 
-let remove_generic_base ctx t = match t with
+let remove_generic_base t = match t with
 	| TClassDecl c when is_removable_class c ->
 		add_class_flag c CExtern;
 	| _ ->
@@ -353,16 +363,16 @@ let remove_generic_base ctx t = match t with
 
 (* Removes extern and macro fields, also checks for Void fields *)
 
-let remove_extern_fields ctx t = match t with
+let remove_extern_fields com t = match t with
 	| TClassDecl c ->
-		if not (Common.defined ctx.com Define.DocGen) then begin
+		if not (Common.defined com Define.DocGen) then begin
 			c.cl_ordered_fields <- List.filter (fun f ->
-				let b = is_removable_field ctx f in
+				let b = is_removable_field com f in
 				if b then c.cl_fields <- PMap.remove f.cf_name c.cl_fields;
 				not b
 			) c.cl_ordered_fields;
 			c.cl_ordered_statics <- List.filter (fun f ->
-				let b = is_removable_field ctx f in
+				let b = is_removable_field com f in
 				if b then c.cl_statics <- PMap.remove f.cf_name c.cl_statics;
 				not b
 			) c.cl_ordered_statics;
@@ -378,12 +388,12 @@ let remove_extern_fields ctx t = match t with
 let check_private_path ctx t = match t with
 	| TClassDecl c when c.cl_private ->
 		let rpath = (fst c.cl_module.m_path,"_" ^ snd c.cl_module.m_path) in
-		if Hashtbl.mem ctx.g.types_module rpath then error ("This private class name will clash with " ^ s_type_path rpath) c.cl_pos;
+		if ctx.com.type_to_module#mem rpath then typing_error ("This private class name will clash with " ^ s_type_path rpath) c.cl_pos;
 	| _ ->
 		()
 
 (* Rewrites class or enum paths if @:native metadata is set *)
-let apply_native_paths ctx t =
+let apply_native_paths t =
 	let get_real_name meta name =
 		let name',p = get_native_name meta in
 		(Meta.RealPath,[Ast.EConst (Ast.String (name,SDoubleQuotes)), p], p), name'
@@ -448,14 +458,14 @@ let apply_native_paths ctx t =
 		()
 
 (* Adds the __rtti field if required *)
-let add_rtti ctx t =
+let add_rtti com t =
 	let rec has_rtti c =
 		Meta.has Meta.Rtti c.cl_meta || match c.cl_super with None -> false | Some (csup,_) -> has_rtti csup
 	in
 	match t with
 	| TClassDecl c when has_rtti c && not (PMap.mem "__rtti" c.cl_statics) ->
-		let f = mk_field ~static:true "__rtti" ctx.t.tstring c.cl_pos null_pos in
-		let str = Genxml.gen_type_string ctx.com t in
+		let f = mk_field ~static:true "__rtti" com.basic.tstring c.cl_pos null_pos in
+		let str = Genxml.gen_type_string com t in
 		f.cf_expr <- Some (mk (TConst (TString str)) f.cf_type c.cl_pos);
 		c.cl_ordered_statics <- f :: c.cl_ordered_statics;
 		c.cl_statics <- PMap.add f.cf_name f c.cl_statics;
@@ -463,9 +473,9 @@ let add_rtti ctx t =
 		()
 
 (* Adds member field initializations as assignments to the constructor *)
-let add_field_inits locals ctx t =
+let add_field_inits cl_path locals com t =
 	let apply c =
-		let ethis = mk (TConst TThis) (TInst (c,List.map snd c.cl_params)) c.cl_pos in
+		let ethis = mk (TConst TThis) (TInst (c,extract_param_types c.cl_params)) c.cl_pos in
 		(* TODO: we have to find a variable name which is not used in any of the functions *)
 		let v = alloc_var VGenerated "_g" ethis.etype ethis.epos in
 		let need_this = ref false in
@@ -482,18 +492,18 @@ let add_field_inits locals ctx t =
 				match cf.cf_expr with
 				| None -> die "" __LOC__
 				| Some e ->
-					let lhs = mk (TField({ ethis with epos = cf.cf_pos },FInstance (c,List.map snd c.cl_params,cf))) cf.cf_type cf.cf_pos in
+					let lhs = mk (TField({ ethis with epos = cf.cf_pos },FInstance (c,extract_param_types c.cl_params,cf))) cf.cf_type cf.cf_pos in
 					cf.cf_expr <- None;
 					mk (TBinop(OpAssign,lhs,e)) cf.cf_type e.epos
 			) inits in
 			let el = if !need_this then (mk (TVar((v, Some ethis))) ethis.etype ethis.epos) :: el else el in
 			let cf = match c.cl_constructor with
 			| None ->
-				let ct = TFun([],ctx.com.basic.tvoid) in
+				let ct = TFun([],com.basic.tvoid) in
 				let ce = mk (TFunction {
 					tf_args = [];
-					tf_type = ctx.com.basic.tvoid;
-					tf_expr = mk (TBlock el) ctx.com.basic.tvoid c.cl_pos;
+					tf_type = com.basic.tvoid;
+					tf_expr = mk (TBlock el) com.basic.tvoid c.cl_pos;
 				}) ct c.cl_pos in
 				let ctor = mk_field "new" ct c.cl_pos null_pos in
 				ctor.cf_kind <- Method MethNormal;
@@ -502,18 +512,20 @@ let add_field_inits locals ctx t =
 				match cf.cf_expr with
 				| Some { eexpr = TFunction f } ->
 					let bl = match f.tf_expr with {eexpr = TBlock b } -> b | x -> [x] in
-					let ce = mk (TFunction {f with tf_expr = mk (TBlock (el @ bl)) ctx.com.basic.tvoid c.cl_pos }) cf.cf_type cf.cf_pos in
+					let ce = mk (TFunction {f with tf_expr = mk (TBlock (el @ bl)) com.basic.tvoid c.cl_pos }) cf.cf_type cf.cf_pos in
 					{cf with cf_expr = Some ce };
 				| _ ->
 					die "" __LOC__
 			in
-			let config = AnalyzerConfig.get_field_config ctx.com c cf in
-			Analyzer.Run.run_on_field ctx config c cf;
+			let config = AnalyzerConfig.get_field_config com c cf in
+			remove_class_field_flag cf CfPostProcessed;
+			Analyzer.Run.run_on_field com config c cf;
+			add_class_field_flag cf CfPostProcessed;
 			(match cf.cf_expr with
 			| Some e ->
 				(* This seems a bit expensive, but hopefully constructor expressions aren't that massive. *)
-				let e = RenameVars.run ctx locals e in
-				let e = Optimizer.sanitize ctx.com e in
+				let e = RenameVars.run cl_path locals e in
+				let e = Optimizer.sanitize com e in
 				cf.cf_expr <- Some e
 			| _ ->
 				());
@@ -526,15 +538,15 @@ let add_field_inits locals ctx t =
 		()
 
 (* Adds the __meta__ field if required *)
-let add_meta_field ctx t = match t with
+let add_meta_field com t = match t with
 	| TClassDecl c ->
-		(match Texpr.build_metadata ctx.com.basic t with
+		(match Texpr.build_metadata com.basic t with
 		| None -> ()
 		| Some e ->
-			add_feature ctx.com "has_metadata";
+			add_feature com "has_metadata";
 			let cf = mk_field ~static:true "__meta__" e.etype e.epos null_pos in
 			cf.cf_expr <- Some e;
-			let can_deal_with_interface_metadata () = match ctx.com.platform with
+			let can_deal_with_interface_metadata () = match com.platform with
 				| Cs | Java -> false
 				| _ -> true
 			in
@@ -544,7 +556,7 @@ let add_meta_field ctx t = match t with
 				let ncls = mk_class c.cl_module path c.cl_pos null_pos in
 				ncls.cl_ordered_statics <- cf :: ncls.cl_ordered_statics;
 				ncls.cl_statics <- PMap.add cf.cf_name cf ncls.cl_statics;
-				ctx.com.types <- ctx.com.types @ [ TClassDecl ncls ];
+				com.types <- com.types @ [ TClassDecl ncls ];
 				c.cl_meta <- (Meta.Custom ":hasMetadata",[],e.epos) :: c.cl_meta
 			end else begin
 				c.cl_ordered_statics <- cf :: c.cl_ordered_statics;
@@ -562,8 +574,8 @@ let check_cs_events com t = match t with
 	| TClassDecl cl when not (has_class_flag cl CExtern) ->
 		let check fields f =
 			match f.cf_kind with
-			| Var { v_read = AccNormal; v_write = AccNormal } when Meta.has Meta.Event f.cf_meta ->
-				if (has_class_field_flag f CfPublic) then error "@:event fields must be private" f.cf_pos;
+			| Var { v_read = AccNormal; v_write = AccNormal } when Meta.has Meta.Event f.cf_meta && not (has_class_field_flag f CfPostProcessed) ->
+				if (has_class_field_flag f CfPublic) then typing_error "@:event fields must be private" f.cf_pos;
 
 				(* prevent generating reflect helpers for the event in gencommon *)
 				f.cf_meta <- (Meta.SkipReflection, [], f.cf_pos) :: f.cf_meta;
@@ -572,7 +584,7 @@ let check_cs_events com t = match t with
 				let tmeth = (tfun [f.cf_type] com.basic.tvoid) in
 
 				let process_event_method name =
-					let m = try PMap.find name fields with Not_found -> error ("Missing event method: " ^ name) f.cf_pos in
+					let m = try PMap.find name fields with Not_found -> typing_error ("Missing event method: " ^ name) f.cf_pos in
 
 					(* check method signature *)
 					begin
@@ -603,17 +615,17 @@ let check_cs_events com t = match t with
 		()
 
 (* Removes interfaces tagged with @:remove metadata *)
-let check_remove_metadata ctx t = match t with
+let check_remove_metadata t = match t with
 	| TClassDecl c ->
 		c.cl_implements <- List.filter (fun (c,_) -> not (Meta.has Meta.Remove c.cl_meta)) c.cl_implements;
 	| _ ->
 		()
 
 (* Checks for Void class fields *)
-let check_void_field ctx t = match t with
+let check_void_field t = match t with
 	| TClassDecl c ->
 		let check f =
-			match follow f.cf_type with TAbstract({a_path=[],"Void"},_) -> error "Fields of type Void are not allowed" f.cf_pos | _ -> ();
+			match follow f.cf_type with TAbstract({a_path=[],"Void"},_) -> typing_error "Fields of type Void are not allowed" f.cf_pos | _ -> ();
 		in
 		List.iter check c.cl_ordered_fields;
 		List.iter check c.cl_ordered_statics;
@@ -623,7 +635,7 @@ let check_void_field ctx t = match t with
 (* Interfaces have no 'super', but can extend many other interfaces.
    This makes the first extended (implemented) interface the super for efficiency reasons (you can get one for 'free')
    and leaves the remaining ones as 'implemented' *)
-let promote_first_interface_to_super ctx t = match t with
+let promote_first_interface_to_super t = match t with
 	| TClassDecl c when (has_class_flag c CInterface) ->
 		begin match c.cl_implements with
 		| ({ cl_path = ["cpp";"rtti"],_ },_ ) :: _ -> ()
@@ -635,16 +647,17 @@ let promote_first_interface_to_super ctx t = match t with
 	| _ ->
 		()
 
-let commit_features ctx t =
+let commit_features com t =
 	let m = (t_infos t).mt_module in
 	Hashtbl.iter (fun k v ->
-		Common.add_feature ctx.com k;
+		Common.add_feature com k;
 	) m.m_extra.m_features
 
 let check_reserved_type_paths ctx t =
 	let check path pos =
-		if List.mem path ctx.com.config.pf_reserved_type_paths then
-			ctx.com.warning ("Type path " ^ (s_type_path path) ^ " is reserved on this target") pos
+		if List.mem path ctx.com.config.pf_reserved_type_paths then begin
+			warning ctx WReservedTypePath ("Type path " ^ (s_type_path path) ^ " is reserved on this target") pos
+		end
 	in
 	match t with
 	| TClassDecl c when not (has_class_flag c CExtern) -> check c.cl_path c.cl_pos
@@ -653,18 +666,12 @@ let check_reserved_type_paths ctx t =
 
 (* PASS 3 end *)
 
-let pp_counter = ref 1
-
-let is_cached t =
+let is_cached com t =
 	let m = (t_infos t).mt_module.m_extra in
-	if m.m_processed = 0 then m.m_processed <- !pp_counter;
-	m.m_processed <> !pp_counter
+	m.m_processed <> 0 && m.m_processed < com.compilation_step
 
 let apply_filters_once ctx filters t =
-	if not (is_cached t) then run_expression_filters None ctx filters t
-
-let next_compilation() =
-	incr pp_counter
+	if not (is_cached ctx.com t) then run_expression_filters None ctx filters t
 
 let iter_expressions fl mt =
 	match mt with
@@ -706,10 +713,212 @@ module ForRemap = struct
 		loop e
 end
 
+let destruction tctx detail_times main locals =
+	let com = tctx.com in
+	let t = filter_timer detail_times ["type 2"] in
+	(* PASS 2: type filters pre-DCE *)
+	List.iter (fun t ->
+		remove_generic_base t;
+		remove_extern_fields com t;
+		(* check @:remove metadata before DCE so it is ignored there (issue #2923) *)
+		check_remove_metadata t;
+	) com.types;
+	t();
+	com.stage <- CDceStart;
+	let t = filter_timer detail_times ["dce"] in
+	(* DCE *)
+	let dce_mode = try Common.defined_value com Define.Dce with _ -> "no" in
+	let dce_mode = match dce_mode with
+		| "full" -> if Common.defined com Define.Interp then Dce.DceNo else DceFull
+		| "std" -> DceStd
+		| "no" -> DceNo
+		| _ -> failwith ("Unknown DCE mode " ^ dce_mode)
+	in
+	Dce.run com main dce_mode;
+	t();
+	com.stage <- CDceDone;
+	(* PASS 3: type filters post-DCE *)
+	List.iter
+		(run_expression_filters
+			~ignore_processed_status:true
+			(timer_label detail_times [])
+			tctx
+			(* This has to run after DCE, or otherwise its condition always holds. *)
+			["insert_save_stacks",Exceptions.insert_save_stacks tctx]
+		)
+		com.types;
+	let type_filters = [
+		Exceptions.patch_constructors tctx; (* TODO: I don't believe this should load_instance anything at this point... *)
+		check_private_path tctx;
+		apply_native_paths;
+		add_rtti com;
+		(match com.platform with | Java | Cs -> (fun _ -> ()) | _ -> (fun mt -> add_field_inits tctx.curclass.cl_path locals com mt));
+		(match com.platform with Hl -> (fun _ -> ()) | _ -> add_meta_field com);
+		check_void_field;
+		(match com.platform with | Cpp -> promote_first_interface_to_super | _ -> (fun _ -> ()));
+		commit_features com;
+		(if com.config.pf_reserved_type_paths <> [] then check_reserved_type_paths tctx else (fun _ -> ()));
+	] in
+	let type_filters = match com.platform with
+		| Cs -> type_filters @ [ fun t -> InterfaceProps.run t ]
+		| _ -> type_filters
+	in
+	let t = filter_timer detail_times ["type 3"] in
+	List.iter (fun t ->
+		begin match t with
+		| TClassDecl c ->
+			tctx.curclass <- c
+		| _ ->
+			()
+		end;
+		List.iter (fun f -> f t) type_filters
+	) com.types;
+	t();
+	com.callbacks#run com.callbacks#get_after_filters;
+	com.stage <- CFilteringDone
+
+let update_cache_dependencies com t =
+	let visited_anons = ref [] in
+	let rec check_t m t = match t with
+		| TInst(c,tl) ->
+			add_dependency m c.cl_module;
+			List.iter (check_t m) tl;
+		| TEnum(en,tl) ->
+			add_dependency m en.e_module;
+			List.iter (check_t m) tl;
+		| TType(t,tl) ->
+			add_dependency m t.t_module;
+			List.iter (check_t m) tl;
+		| TAbstract(a,tl) ->
+			add_dependency m a.a_module;
+			List.iter (check_t m) tl;
+		| TFun(targs,tret) ->
+			List.iter (fun (_,_,t) -> check_t m t) targs;
+			check_t m tret;
+		| TAnon an ->
+			if not (List.memq an !visited_anons) then begin
+				visited_anons := an :: !visited_anons;
+				PMap.iter (fun _ cf -> check_t m cf.cf_type) an.a_fields
+			end
+		| TMono r ->
+			begin match r.tm_type with
+				| Some t ->
+					check_t m t
+				| _ ->
+					(* Bind any still open monomorph that's part of a signature to Dynamic now (issue #10653) *)
+					Monomorph.do_bind r t_dynamic;
+		end
+		| TLazy f ->
+			check_t m (lazy_type f)
+		| TDynamic None ->
+			()
+		| TDynamic (Some t) ->
+			check_t m t
+	in
+	let rec check_field m cf =
+		check_t m cf.cf_type;
+		List.iter (check_field m) cf.cf_overloads
+	in
+	match t with
+		| TClassDecl c ->
+			List.iter (check_field c.cl_module) c.cl_ordered_statics;
+			List.iter (check_field c.cl_module) c.cl_ordered_fields;
+			(match c.cl_constructor with None -> () | Some cf -> check_field c.cl_module cf);
+		| _ ->
+			()
+
+(* Saves a class state so it can be restored later, e.g. after DCE or native path rewrite *)
+let save_class_state ctx t =
+	(* Update m_processed here. This means that nothing should add a dependency afterwards because
+	   then the module is immediately considered uncached again *)
+	(t_infos t).mt_module.m_extra.m_processed <- ctx.com.compilation_step;
+	match t with
+	| TClassDecl c ->
+		let vars = ref [] in
+		let rec save_vars e =
+			let add v = vars := (v, v.v_type) :: !vars in
+			match e.eexpr with
+				| TFunction fn ->
+					List.iter (fun (v, _) -> add v) fn.tf_args;
+					save_vars fn.tf_expr
+				| TVar (v, e) ->
+					add v;
+					Option.may save_vars e
+				| _ ->
+					iter save_vars e
+		in
+		let mk_field_restore f =
+			Option.may save_vars f.cf_expr;
+			let rec mk_overload_restore f =
+				add_class_field_flag f CfPostProcessed;
+				f.cf_name,f.cf_kind,f.cf_expr,f.cf_type,f.cf_meta,f.cf_params
+			in
+			( f,mk_overload_restore f, List.map (fun f -> f,mk_overload_restore f) f.cf_overloads )
+		in
+		let restore_field (f,res,overloads) =
+			let restore_field (f,(name,kind,expr,t,meta,params)) =
+				f.cf_name <- name; f.cf_kind <- kind; f.cf_expr <- expr; f.cf_type <- t; f.cf_meta <- meta; f.cf_params <- params;
+				f
+			in
+			let f = restore_field (f,res) in
+			f.cf_overloads <- List.map restore_field overloads;
+			f
+		in
+		let mk_pmap lst =
+			List.fold_left (fun pmap f -> PMap.add f.cf_name f pmap) PMap.empty lst
+		in
+
+		let meta = c.cl_meta and path = c.cl_path and ext = (has_class_flag c CExtern) in
+		let sup = c.cl_super and impl = c.cl_implements in
+		let csr = Option.map (mk_field_restore) c.cl_constructor in
+		let ofr = List.map (mk_field_restore) c.cl_ordered_fields in
+		let osr = List.map (mk_field_restore) c.cl_ordered_statics in
+		let init = c.cl_init in
+		Option.may save_vars init;
+		c.cl_restore <- (fun() ->
+			c.cl_super <- sup;
+			c.cl_implements <- impl;
+			c.cl_meta <- meta;
+			if ext then add_class_flag c CExtern else remove_class_flag c CExtern;
+			c.cl_path <- path;
+			c.cl_init <- init;
+			c.cl_ordered_fields <- List.map restore_field ofr;
+			c.cl_ordered_statics <- List.map restore_field osr;
+			c.cl_fields <- mk_pmap c.cl_ordered_fields;
+			c.cl_statics <- mk_pmap c.cl_ordered_statics;
+			c.cl_constructor <- Option.map restore_field csr;
+			c.cl_descendants <- [];
+			List.iter (fun (v, t) -> v.v_type <- t) !vars;
+		)
+	| TEnumDecl en ->
+		let path = en.e_path in
+		en.e_restore <- (fun () ->
+			let rec loop acc = function
+				| [] ->
+					en.e_path <- path;
+				| (Meta.RealPath,[Ast.EConst (Ast.String(path,_)),_],_) :: l ->
+					en.e_path <- Ast.parse_path path;
+					en.e_meta <- (List.rev acc) @ l;
+				| x :: l -> loop (x::acc) l
+			in
+			loop [] en.e_meta
+		)
+	| TTypeDecl td ->
+		let path = td.t_path in
+		td.t_restore <- (fun () ->
+			td.t_path <- path
+		);
+	| TAbstractDecl a ->
+		let path = a.a_path in
+		a.a_restore <- (fun () ->
+			a.a_path <- path;
+			a.a_meta <- List.filter (fun (m,_,_) -> m <> Meta.ValueUsed) a.a_meta
+		)
+
 let run com tctx main =
 	let detail_times = Common.defined com DefineList.FilterTimes in
 	let new_types = List.filter (fun t ->
-		let cached = is_cached t in
+		let cached = is_cached com t in
 		begin match t with
 			| TClassDecl cls ->
 				List.iter (fun (iface,_) -> add_descendant iface cls) cls.cl_implements;
@@ -720,9 +929,8 @@ let run com tctx main =
 				(* Save cf_expr_unoptimized early: We want to inline with the original expression
 				   on the next compilation. *)
 				if not cached then begin
-					let field cf = match cf.cf_expr with
-						| Some {eexpr = TFunction tf} -> cf.cf_expr_unoptimized <- Some tf
-						| _ -> ()
+					let field cf =
+						cf.cf_expr_unoptimized <- cf.cf_expr
 					in
 					List.iter field cls.cl_ordered_fields;
 					List.iter field cls.cl_ordered_statics;
@@ -732,6 +940,15 @@ let run com tctx main =
 		end;
 		not cached
 	) com.types in
+	(* IMPORTANT:
+	    There may be types in new_types which have already been post-processed, but then had their m_processed flag unset
+		because they received an additional dependency. This could happen in cases such as @:generic methods in #10635.
+		It is important that all filters from here up to save_class_state only process fields which do not have the
+		CfPostProcessed flag set.
+
+		This is mostly covered by run_expression_filters already, but any new additions which don't utilize that have to
+		be aware of this.
+	*)
 	NullSafety.run com new_types;
 	(* PASS 1: general expression filters *)
 	let filters = [
@@ -741,8 +958,9 @@ let run com tctx main =
 	] in
 	List.iter (run_expression_filters (timer_label detail_times ["expr 0"]) tctx filters) new_types;
 	let filters = [
+		"local_statics",LocalStatic.run tctx;
 		"fix_return_dynamic_from_void_function",fix_return_dynamic_from_void_function tctx true;
-		"check_local_vars_init",check_local_vars_init tctx.com;
+		"check_local_vars_init",check_local_vars_init tctx;
 		"check_abstract_as_value",check_abstract_as_value;
 		"Tre",if defined com Define.AnalyzerOptimize then Tre.run tctx else (fun e -> e);
 		"reduce_expression",Optimizer.reduce_expression tctx;
@@ -780,7 +998,7 @@ let run com tctx main =
 	List.iter (fun f -> List.iter f new_types) filters;
 	t();
 	com.stage <- CAnalyzerStart;
-	if com.platform <> Cross then Analyzer.Run.run_on_types tctx new_types;
+	if com.platform <> Cross then Analyzer.Run.run_on_types com new_types;
 	com.stage <- CAnalyzerDone;
 	let locals = RenameVars.init com in
 	let filters = [
@@ -788,71 +1006,22 @@ let run com tctx main =
 		"add_final_return",if com.config.pf_add_final_return then add_final_return else (fun e -> e);
 		"RenameVars",(match com.platform with
 		| Eval -> (fun e -> e)
-		| _ -> RenameVars.run tctx locals);
+		| _ -> (fun e -> RenameVars.run tctx.curclass.cl_path locals e));
 		"mark_switch_break_loops",mark_switch_break_loops;
 	] in
 	List.iter (run_expression_filters (timer_label detail_times ["expr 2"]) tctx filters) new_types;
-	next_compilation();
 	let t = filter_timer detail_times ["callbacks"] in
-	List.iter (fun f -> f()) (List.rev com.callbacks#get_before_save); (* macros onGenerate etc. *)
+	com.callbacks#run com.callbacks#get_before_save; (* macros onGenerate etc. *)
 	t();
 	com.stage <- CSaveStart;
 	let t = filter_timer detail_times ["save state"] in
-	List.iter (save_class_state tctx) new_types;
+	List.iter (fun mt ->
+		update_cache_dependencies com mt;
+		save_class_state tctx mt
+	) new_types;
 	t();
 	com.stage <- CSaveDone;
 	let t = filter_timer detail_times ["callbacks"] in
-	List.iter (fun f -> f()) (List.rev com.callbacks#get_after_save); (* macros onGenerate etc. *)
+	com.callbacks#run com.callbacks#get_after_save; (* macros onGenerate etc. *)
 	t();
-	let t = filter_timer detail_times ["type 2"] in
-	(* PASS 2: type filters pre-DCE *)
-	List.iter (fun t ->
-		remove_generic_base tctx t;
-		remove_extern_fields tctx t;
-		Codegen.update_cache_dependencies t;
-		(* check @:remove metadata before DCE so it is ignored there (issue #2923) *)
-		check_remove_metadata tctx t;
-	) com.types;
-	t();
-	com.stage <- CDceStart;
-	let t = filter_timer detail_times ["dce"] in
-	(* DCE *)
-	let dce_mode = try Common.defined_value com Define.Dce with _ -> "no" in
-	let dce_mode = match dce_mode with
-		| "full" -> if Common.defined com Define.Interp then Dce.DceNo else DceFull
-		| "std" -> DceStd
-		| "no" -> DceNo
-		| _ -> failwith ("Unknown DCE mode " ^ dce_mode)
-	in
-	Dce.run com main dce_mode;
-	t();
-	com.stage <- CDceDone;
-	(* PASS 3: type filters post-DCE *)
-	List.iter
-		(run_expression_filters
-			(timer_label detail_times [])
-			tctx
-			["insert_save_stacks",Exceptions.insert_save_stacks tctx]
-		)
-		new_types;
-	let type_filters = [
-		Exceptions.patch_constructors;
-		check_private_path;
-		apply_native_paths;
-		add_rtti;
-		(match com.platform with | Java | Cs -> (fun _ _ -> ()) | _ -> add_field_inits locals);
-		(match com.platform with Hl -> (fun _ _ -> ()) | _ -> add_meta_field);
-		check_void_field;
-		(match com.platform with | Cpp -> promote_first_interface_to_super | _ -> (fun _ _ -> ()) );
-		commit_features;
-		(if com.config.pf_reserved_type_paths <> [] then check_reserved_type_paths else (fun _ _ -> ()));
-	] in
-	let type_filters = match com.platform with
-		| Cs -> type_filters @ [ fun _ t -> InterfaceProps.run t ]
-		| _ -> type_filters
-	in
-	let t = filter_timer detail_times ["type 3"] in
-	List.iter (fun t -> List.iter (fun f -> f tctx t) type_filters) com.types;
-	t();
-	List.iter (fun f -> f()) (List.rev com.callbacks#get_after_filters);
-	com.stage <- CFilteringDone
+	destruction tctx detail_times main locals
\ No newline at end of file
