@@ -78,6 +78,7 @@ type context = {
 	mutable file_prefix : string;
 	mutable fun_index : int;
 	mutable type_module : (ttype, code_module) PMap.t;
+	gcon : Common.context;
 }
 
 let sprintf = Printf.sprintf
@@ -124,7 +125,7 @@ let tname str =
 	if Hashtbl.mem keywords ("_" ^ n) then "__" ^ n else n
 
 let is_gc_ptr = function
-	| HVoid | HUI8 | HUI16 | HI32 | HI64 | HF32 | HF64 | HBool | HType | HRef _ | HMethod _ -> false
+	| HVoid | HUI8 | HUI16 | HI32 | HI64 | HF32 | HF64 | HBool | HType | HRef _ | HMethod _ | HPacked _ -> false
 	| HBytes | HDyn | HFun _ | HObj _ | HArray | HVirtual _ | HDynObj | HAbstract _ | HEnum _ | HNull _ | HStruct _ -> true
 
 let is_ptr = function
@@ -153,6 +154,9 @@ let rec ctype_no_ptr = function
 	| HEnum _ -> "venum",1
 	| HNull _ -> "vdynamic",1
 	| HMethod _ -> "void",1
+	| HPacked t ->
+		let name,v = ctype_no_ptr t in
+		"struct _" ^ name, v
 
 let ctype t =
 	let t, nptr = ctype_no_ptr t in
@@ -199,6 +203,7 @@ let type_id t =
 	| HNull _ -> "HNULL"
 	| HMethod _ -> "HMETHOD"
 	| HStruct _  -> "HSTRUCT"
+	| HPacked _ -> "HPACKED"
 
 let var_type n t =
 	ctype t ^ " " ^ ident n
@@ -223,7 +228,7 @@ let hash ctx sid =
 		h
 
 let type_name ctx t =
-	try PMap.find t ctx.htypes with Not_found -> Globals.die "" __LOC__
+	try PMap.find t ctx.htypes with Not_found -> Globals.die (tstr t) __LOC__
 
 let define ctx s =
 	if not (Hashtbl.mem ctx.hdefines s) then begin
@@ -245,6 +250,8 @@ let rec define_type ctx t =
 	| HVirtual vp when not (PMap.exists t ctx.defined_types) ->
 		ctx.defined_types <- PMap.add t () ctx.defined_types;
 		Array.iter (fun (_,_,t) -> define_type ctx t) vp.vfields
+	| HPacked t ->
+		define_type ctx t
 	| HEnum _ | HObj _ | HStruct _ | HVirtual _ ->
 		()
 
@@ -291,10 +298,13 @@ let unamed_field fid = "f$" ^ string_of_int fid
 let obj_field fid name =
 	if name = "" then unamed_field fid else ident name
 
+let bom = "\xEF\xBB\xBF"
+
 let close_file ctx =
-	let str = Buffer.contents ctx.out in
+	let out = Buffer.contents ctx.out in
 	let defines = List.rev ctx.defines in
-	let str = (match defines with [] -> str | l -> String.concat "\n" l ^ "\n\n" ^ str) in
+	let content = (match defines with [] -> out | l -> String.concat "\n" l ^ "\n\n" ^ out) in
+	let str = if ctx.curfile = "hlc.json" then content else bom ^ content in
 	ctx.defines <- [];
 	ctx.defined_types <- PMap.empty;
 	Hashtbl.clear ctx.hdefines;
@@ -311,8 +321,6 @@ let close_file ctx =
 		close_out ch;
 	end
 
-let bom = "\xEF\xBB\xBF"
-
 let define_global ctx g =
 	let t = ctx.hlcode.globals.(g) in
 	define_type ctx t;
@@ -337,10 +345,8 @@ let short_digest str =
 
 let open_file ctx file =
 	if ctx.curfile <> "" then close_file ctx;
-	let version_major = ctx.version / 1000 in
-	let version_minor = (ctx.version mod 1000) / 100 in
-	let version_revision = (ctx.version mod 100) in
-	if file <> "hlc.json" then define ctx (sprintf "%s// Generated by HLC %d.%d.%d (HL v%d)" bom version_major version_minor version_revision ctx.hlcode.version);
+	if file <> "hlc.json" then
+		Codegen.map_source_header ctx.gcon (fun s -> define ctx (sprintf "// %s" s));
 	ctx.curfile <- file;
 	ctx.fun_index <- 0;
 	ctx.file_prefix <- (short_digest file) ^ "_"
@@ -549,7 +555,9 @@ let generate_function ctx f =
 		let assign = reg r ^ " = " in
 		if tsame t rt then assign else
 		if not (safe_cast t rt) then Globals.die "" __LOC__
-		else assign ^ "(" ^ ctype rt ^ ")"
+		else
+			let cast = assign ^ "(" ^ ctype rt ^ ")" in
+			(match t with HPacked _ -> cast ^ "&" | _ -> cast)
 	in
 
 	let ocall r fid args =
@@ -568,7 +576,7 @@ let generate_function ctx f =
 	in
 
 	let type_value_opt t =
-		match t with HF32 | HF64 -> "" | _ -> "," ^ type_value t
+		match t with HF32 | HF64 | HI64 -> "" | _ -> "," ^ type_value t
 	in
 
 	let dyn_call r f pl =
@@ -735,7 +743,7 @@ let generate_function ctx f =
 				one way for comparisons
 			*)
 			match rtype a, rtype b with
-			| (HUI8 | HUI16 | HI32 | HF32 | HF64 | HBool), (HUI8 | HUI16 | HI32 | HF32 | HF64 | HBool) ->
+			| (HUI8 | HUI16 | HI32 | HF32 | HF64 | HBool | HI64), (HUI8 | HUI16 | HI32 | HF32 | HF64 | HBool | HI64) ->
 				phys_compare()
 			| HType, HType ->
 				sexpr "if( hl_same_type(%s,%s) %s 0 ) {} else goto %s" (reg a) (reg b) (s_comp op) (label d)
@@ -1096,6 +1104,9 @@ let valid_ident =
 	let e = Str.regexp "[^A-Za-z0-9_]+" in
 	(fun str -> Str.global_replace e "_" str)
 
+let native_name str =
+	if str.[0] = '?' then String.sub str 1 (String.length str - 1) else str
+
 let make_types_idents htypes =
 	let types_descs = ref PMap.empty in
 	let rec make_desc t =
@@ -1108,6 +1119,10 @@ let make_types_idents htypes =
 			DFun (List.map make_desc tl, make_desc t, false)
 		| HObj p | HStruct p ->
 			DNamed p.pname
+		| HPacked t ->
+			(match make_desc t with
+			| DNamed n -> DNamed ("packed_" ^ n)
+			| _ -> Globals.die "" __LOC__)
 		| HAbstract (n,_) ->
 			DNamed n
 		| HEnum e when e.ename = "" ->
@@ -1125,8 +1140,13 @@ let make_types_idents htypes =
 				Array.iteri (fun i (f,_,t) -> arr.(i) <- (f,make_desc t)) vp.vfields;
 				td
 	in
+	let hashes = Hashtbl.create 0 in
 	let make_sign d =
-		String.sub (Digest.to_hex (Digest.bytes (Marshal.to_bytes d [Marshal.Compat_32]))) 0 7
+		let dig = Digest.to_hex (Digest.bytes (Marshal.to_bytes d [Marshal.Compat_32])) in
+		let h = String.sub dig 0 7 in
+		let h = if Hashtbl.mem hashes h then dig else h in
+		Hashtbl.add hashes h ();
+		h
 	in
 	let rec desc_string d =
 		match d with
@@ -1182,8 +1202,11 @@ let make_function_table code =
 	Array.iter (fun (lib,name,t,idx) ->
 		let fname =
 			let lib = code.strings.(lib) in
-			let lib = if lib = "std" then "hl" else lib in
-			lib ^ "_" ^ code.strings.(name)
+			let lib = if lib.[0] = '?' then String.sub lib 1 (String.length lib - 1) else lib in
+ 			let lib = if lib = "std" then "hl" else lib in
+			let str = lib ^ "_" ^ code.strings.(name) in
+			(* create wrappers defines for invalid definitions *)
+			if str = "hl_tls_get" then str ^ "_w" else str
 		in
 		match t with
 		| HFun (args, t) ->
@@ -1430,6 +1453,7 @@ let write_c com file (code:code) gnames =
 		file_prefix = "";
 		fun_index = 0;
 		type_module = PMap.empty;
+		gcon = com;
 	} in
 	let modules = make_modules ctx all_types in
 
@@ -1442,6 +1466,7 @@ let write_c com file (code:code) gnames =
 	define ctx "// Abstract decls";
 	let rec get_abstracts = function
 		| [] -> []
+		| HAbstract (("hl_tls" | "hl_mutex" | "hl_thread"), _) :: l -> get_abstracts l (* ignore / defined in hl.h already *)
 		| HAbstract (name,_) :: l -> name :: get_abstracts l
 		| _ :: l -> get_abstracts l
 	in
@@ -1453,10 +1478,18 @@ let write_c com file (code:code) gnames =
 	let sorted_natives = Array.copy code.natives in
 	Array.sort (fun n1 n2 -> let mk (lib,name,_,_) = code.strings.(lib), code.strings.(name) in compare (mk n1) (mk n2)) sorted_natives;
 	Array.iter (fun (lib,_,_,idx) ->
-		Hashtbl.replace native_libs code.strings.(lib) ();
+		let name = code.strings.(lib) in
+		let name = if name.[0] = '?' then String.sub name 1 (String.length name - 1) else name in
+		Hashtbl.replace native_libs name ();
 		let ft = ctx.ftable.(idx) in
 		define_type ctx (HFun (ft.fe_args,ft.fe_ret));
-		sexpr "HL_API %s %s(%s)" (ctype ft.fe_ret) ft.fe_name (args_repr ft.fe_args);
+		match ft.fe_name with
+		| "hl_tls_get_w" ->
+			define ctx "#define hl_tls_get_w(tls) ((vdynamic*)hl_tls_get(tls))";
+		| "hl_tls_set" ->
+			() (* don't redefine *)
+		| _ ->
+			sexpr "HL_API %s %s(%s)" (ctype ft.fe_ret) ft.fe_name (args_repr ft.fe_args);
 	) sorted_natives;
 	line "#endif";
 	line "";
@@ -1588,7 +1621,7 @@ let write_c com file (code:code) gnames =
 				string_of_int (Array.length o.pproto);
 				string_of_int (List.length o.pbindings);
 				sprintf "(const uchar*)%s" (string ctx o.pid);
-				(match o.psuper with None -> "NULL" | Some c -> type_value ctx (HObj c));
+				(match o.psuper with None -> "NULL" | Some c -> type_value ctx (match t with HObj _ -> HObj c | _ -> HStruct c));
 				fields;
 				proto;
 				bindings
@@ -1661,7 +1694,7 @@ let write_c com file (code:code) gnames =
 				define_global ctx g;
 				sexpr "obj%s.global_value = (void**)&%s" name gnames.(g));
 			sexpr "%s.obj = &obj%s" name name
-		| HNull r | HRef r ->
+		| HNull r | HRef r | HPacked r ->
 			sexpr "%s.tparam = %s" (type_name ctx t) (type_value ctx r)
 		| HEnum e ->
 			let name = type_name ctx t in
@@ -1720,7 +1753,10 @@ let write_c com file (code:code) gnames =
 	open_file ctx "hl/functions.c";
 	define ctx "#define HLC_BOOT";
 	define ctx "#include <hlc.h>";
-	sexpr "void *hl_functions_ptrs[] = {%s}" (String.concat "," (List.map (fun f -> define_function ctx f.fe_index) (Array.to_list ctx.ftable)));
+	sexpr "void *hl_functions_ptrs[] = {%s}" (String.concat ",\\\n\t" (List.map (fun f ->
+		let name = define_function ctx f.fe_index in
+		if name = "hl_tls_get_w" then "hl_tls_get" else name
+	) (Array.to_list ctx.ftable)));
 	let rec loop i =
 		if i = Array.length ctx.ftable then [] else
 		let ft = ctx.ftable.(i) in
@@ -1728,7 +1764,7 @@ let write_c com file (code:code) gnames =
 		define ctx (sprintf "extern hl_type %s;" n);
 		("&" ^ n) :: loop (i + 1)
 	in
-	sexpr "hl_type *hl_functions_types[] = {%s}" (String.concat "," (loop 0));
+	sexpr "hl_type *hl_functions_types[] = {%s}" (String.concat ",\\\n\t" (loop 0));
 	line "";
 	Array.iter (fun f ->
 		if f.fe_module = None then (match f.fe_decl with None -> () | Some f -> generate_function ctx f);
@@ -1780,10 +1816,14 @@ let write_c com file (code:code) gnames =
 	block ctx;
 	sline "\"version\" : %d," ctx.version;
 	sline "\"libs\" : [%s]," (String.concat "," (Hashtbl.fold (fun k _ acc -> sprintf "\"%s\"" k :: acc) native_libs []));
-	sline "\"defines\" : {%s\n\t}," (String.concat "," (PMap.foldi (fun k v acc -> sprintf "\n\t\t\"%s\" : \"%s\"" (String.escaped k) (String.escaped v) :: acc) com.Common.defines.Define.values []));
+	let defines = Buffer.create 64 in
+	PMap.iter (fun key value ->
+		Printf.bprintf defines "\n\t\t\"%s\" : \"%s\"," (String.escaped key) (String.escaped value);
+	) com.defines.values;
+	Buffer.truncate defines (Buffer.length defines - 1);
+	sline "\"defines\" : {%s\n\t}," (Buffer.contents defines);
 	sline "\"files\" : [%s\n\t]" (String.concat "," (List.map (sprintf "\n\t\t\"%s\"") ctx.cfiles));
 	unblock ctx;
 	line "}";
 
 	close_file ctx
-
