@@ -235,7 +235,7 @@ and jit_expr jit return e =
 		List.iter (fun var -> ignore(get_capture_slot jit var)) jit_closure.captures_outside_scope;
 		let captures = ExtList.List.filter_map (fun (i,vid,declared) ->
 			if declared then None
-			else Some (i,fst (try Hashtbl.find jit.captures vid with Not_found -> Error.error "Something went wrong" e.epos))
+			else Some (i,fst (try Hashtbl.find jit.captures vid with Not_found -> Error.typing_error "Something went wrong" e.epos))
 		) captures in
 		let mapping = Array.of_list captures in
 		emit_closure ctx mapping eci hasret exec fl
@@ -429,14 +429,18 @@ and jit_expr jit return e =
 				let i = get_proto_field_index proto name in
 				lazy (match proto.pfields.(i) with VFunction (f,_) -> f | v -> cannot_call v e.epos)
 			in
-			let instance_call c =
+			let jit_with_null_check ef =
 				let exec = jit_expr jit false ef in
+				emit_null_check exec ef.epos
+			in
+			let instance_call c =
+				let exec = jit_with_null_check ef in
 				let proto = get_instance_prototype ctx (path_hash c.cl_path) ef.epos in
 				let v = lazy_proto_field proto in
 				emit_proto_field_call v (exec :: execs) e.epos
 			in
 			let default () =
-				let exec = jit_expr jit false ef in
+				let exec = jit_with_null_check ef in
 				emit_method_call exec name execs e.epos
 			in
 			begin match fa with
@@ -472,7 +476,7 @@ and jit_expr jit return e =
 					end else
 						default()
 				| _ ->
-					let exec = jit_expr jit false ef in
+					let exec = jit_with_null_check ef in
 					emit_field_call exec name execs e.epos
 			end
 		| TConst TSuper ->
@@ -507,7 +511,7 @@ and jit_expr jit return e =
 	| TNew({cl_path=["eval"],"Vector"},_,[e1]) ->
 		begin match e1.eexpr with
 			| TConst (TInt i32) ->
-				emit_new_vector_int (Int32.to_int i32)
+				emit_new_vector_int (Int32.to_int i32) e1.epos
 			| _ ->
 				let exec1 = jit_expr jit false e1 in
 				emit_new_vector exec1 e1.epos
@@ -633,7 +637,7 @@ and jit_expr jit return e =
 	| TParenthesis e1 | TMeta(_,e1) | TCast(e1,None) ->
 		loop e1
 	| TIdent s ->
-		Error.error ("Unknown identifier: " ^ s) e.epos
+		Error.typing_error ("Unknown identifier: " ^ s) e.epos
 	in
 	let f = loop e in
 	begin match ctx.debug.debug_socket with
