@@ -122,8 +122,8 @@ let inline_constructors ctx original_e =
 			| IOKCtor(ioc) ->
 				List.iter (fun v -> if v.v_id < 0 then cancel_v v p) io.io_dependent_vars;
 				if ioc.ioc_forced then begin
-					display_error ctx "Forced inline constructor could not be inlined" io.io_pos;
-					display_error ctx (compl_msg "Cancellation happened here") p;
+					display_error ctx.com "Forced inline constructor could not be inlined" io.io_pos;
+					display_error ~depth:1 ctx.com (compl_msg "Cancellation happened here") p;
 				end
 			| _ -> ()
 		end
@@ -227,7 +227,7 @@ let inline_constructors ctx original_e =
 		let e = Type.map_expr (mark_ctors ~force_inline:is_meta_inline) e in
 		let mark() =
 			incr curr_io_id;
-			let id_expr = (EConst(Int (string_of_int !curr_io_id)), e.epos) in
+			let id_expr = (EConst(Int (string_of_int !curr_io_id, None)), e.epos) in
 			let meta = (Meta.InlineObject, [id_expr], e.epos) in
 			mk (TMeta(meta, e)) e.etype e.epos
 		in
@@ -316,7 +316,11 @@ let inline_constructors ctx original_e =
 						let iv_is_const iv = match iv.iv_kind with IVKField(_,_,Some(_)) -> true | _ -> false in
 						if is_lvalue && iv_is_const fiv then raise Not_found;
 						if fiv.iv_closed then raise Not_found;
-						if not captured || (not is_lvalue && fiv.iv_state == IVSUnassigned) then cancel_iv fiv efield.epos;
+						if not is_lvalue && fiv.iv_state == IVSUnassigned then (
+							warning ctx WConstructorInliningCancelled ("Constructor inlining cancelled because of use of uninitialized member field " ^ fname) ethis.epos;
+							raise Not_found
+						);
+						if not captured then cancel_iv fiv efield.epos;
 						IOFInlineVar(fiv)
 					with Not_found ->
 						cancel_iv iv efield.epos;
@@ -395,11 +399,11 @@ let inline_constructors ctx original_e =
 						Some iv
 					| _ ->
 						List.iter (fun v -> cancel_v v v.v_pos) argvs;
-						if is_extern_ctor c cf then display_error ctx "Extern constructor could not be inlined" e.epos;
+						if is_extern_ctor c cf then display_error ctx.com "Extern constructor could not be inlined" e.epos;
 						None
 				end
 			| TNew({ cl_constructor = Some ({cf_kind = Method MethInline; cf_expr = Some _} as cf)} as c,_,pl),_ when is_extern_ctor c cf ->
-				error "Extern constructor could not be inlined" e.epos;
+				typing_error "Extern constructor could not be inlined" e.epos;
 			| TObjectDecl fl, _ when captured && fl <> [] && List.for_all (fun((s,_,_),_) -> Lexer.is_valid_identifier s) fl ->
 				let v = alloc_var VGenerated "inlobj" e.etype e.epos in
 				let ev = mk (TLocal v) v.v_type e.epos in
@@ -435,10 +439,10 @@ let inline_constructors ctx original_e =
 				handle_default_case e
 		in
 		match e.eexpr with
-		| TMeta((Meta.Inline,_,_),{eexpr = TMeta((Meta.InlineObject, [(EConst(Int (id_str)), _)], _), e)}) ->
+		| TMeta((Meta.Inline,_,_),{eexpr = TMeta((Meta.InlineObject, [(EConst(Int (id_str, None)), _)], _), e)}) ->
 			let io_id = int_of_string id_str in
 			handle_inline_object_case io_id true e
-		| TMeta((Meta.InlineObject, [(EConst(Int (id_str)), _)], _), e) ->
+		| TMeta((Meta.InlineObject, [(EConst(Int (id_str, None)), _)], _), e) ->
 			let io_id = int_of_string id_str in
 			handle_inline_object_case io_id false e
 		| TVar(v,None) -> ignore(add v IVKLocal); None
@@ -528,6 +532,8 @@ let inline_constructors ctx original_e =
 				List.iter (fun ca -> ignore(analyze_aliases false ca)) call_args;
 				None
 			end
+		| TFunction tf ->
+			analyze_aliases true tf.tf_expr
 		| _ ->
 			handle_default_case e
 	in
@@ -588,7 +594,7 @@ let inline_constructors ctx original_e =
 			end
 		in
 		match e.eexpr with
-		| TMeta((Meta.InlineObject, [(EConst(Int (id_str)), _)], _), e) ->
+		| TMeta((Meta.InlineObject, [(EConst(Int (id_str, _)), _)], _), e) ->
 			let io_id = int_of_string id_str in
 			begin try
 				let io = get_io io_id in
