@@ -37,16 +37,17 @@ exception Build_canceled of build_state
 
 let is_generic_parameter ctx c =
 	(* first check field parameters, then class parameters *)
+	let name = snd c.cl_path in
 	try
-		ignore (List.assoc (snd c.cl_path) ctx.curfield.cf_params);
+		ignore(lookup_param name ctx.curfield.cf_params);
 		has_class_field_flag ctx.curfield CfGeneric
 	with Not_found -> try
-		ignore(List.assoc (snd c.cl_path) ctx.type_params);
+		ignore(lookup_param name ctx.type_params);
 		(match ctx.curclass.cl_kind with | KGeneric -> true | _ -> false);
 	with Not_found ->
 		false
 
-let valid_redefinition ctx f1 t1 f2 t2 = (* child, parent *)
+let valid_redefinition ctx map1 map2 f1 t1 f2 t2 = (* child, parent *)
 	let valid t1 t2 =
 		Type.unify t1 t2;
 		if is_null t1 <> is_null t2 || ((follow t1) == t_dynamic && (follow t2) != t_dynamic) then raise (Unify_error [Cannot_unify (t1,t2)]);
@@ -61,8 +62,9 @@ let valid_redefinition ctx f1 t1 f2 t2 = (* child, parent *)
 		| [], [] -> t1, t2
 		| l1, l2 when List.length l1 = List.length l2 ->
 			let to_check = ref [] in
-			let monos = List.map2 (fun (name,p1) (_,p2) ->
-				(match follow p1, follow p2 with
+			(* TPTODO: defaults *)
+			let monos = List.map2 (fun tp1 tp2 ->
+				(match follow tp1.ttp_type, follow tp2.ttp_type with
 				| TInst ({ cl_kind = KTypeParameter ct1 } as c1,pl1), TInst ({ cl_kind = KTypeParameter ct2 } as c2,pl2) ->
 					(match ct1, ct2 with
 					| [], [] -> ()
@@ -71,8 +73,8 @@ let valid_redefinition ctx f1 t1 f2 t2 = (* child, parent *)
 						let check monos =
 							List.iter2 (fun t1 t2  ->
 								try
-									let t1 = apply_params l1 monos (apply_params c1.cl_params pl1 t1) in
-									let t2 = apply_params l2 monos (apply_params c2.cl_params pl2 t2) in
+									let t1 = apply_params l1 monos (apply_params c1.cl_params pl1 (map2 t1)) in
+									let t2 = apply_params l2 monos (apply_params c2.cl_params pl2 (map1 t2)) in
 									type_eq EqStrict t1 t2
 								with Unify_error l ->
 									raise (Unify_error (Unify_custom "Constraints differ" :: l))
@@ -82,7 +84,7 @@ let valid_redefinition ctx f1 t1 f2 t2 = (* child, parent *)
 					| _ ->
 						raise (Unify_error [Unify_custom "Different number of constraints"]))
 				| _ -> ());
-				TInst (mk_class null_module ([],name) null_pos null_pos,[])
+				TInst (mk_class null_module ([],tp1.ttp_name) null_pos null_pos,[])
 			) l1 l2 in
 			List.iter (fun f -> f monos) !to_check;
 			apply_params l1 monos t1, apply_params l2 monos t2
@@ -146,12 +148,12 @@ let get_native_name meta =
 	| [] ->
 		raise Not_found
 	| _ ->
-		error "String expected" mp
+		typing_error "String expected" mp
 
 let check_native_name_override ctx child base =
 	let error base_pos child_pos =
-		display_error ctx ("Field " ^ child.cf_name ^ " has different @:native value than in superclass") child_pos;
-		display_error ctx (compl_msg "Base field is defined here") base_pos
+		display_error ctx.com ("Field " ^ child.cf_name ^ " has different @:native value than in superclass") child_pos;
+		display_error ~depth:1 ctx.com (compl_msg "Base field is defined here") base_pos
 	in
 	try
 		let child_name, child_pos = get_native_name child.cf_meta in
@@ -166,14 +168,14 @@ let check_native_name_override ctx child base =
 let check_overriding ctx c f =
 	match c.cl_super with
 	| None ->
-		if has_class_field_flag f CfOverride then display_error ctx ("Field " ^ f.cf_name ^ " is declared 'override' but doesn't override any field") f.cf_pos
+		if has_class_field_flag f CfOverride then display_error ctx.com ("Field " ^ f.cf_name ^ " is declared 'override' but doesn't override any field") f.cf_pos
 	| _ when (has_class_flag c CExtern) && Meta.has Meta.CsNative c.cl_meta -> () (* -net-lib specific: do not check overrides on extern CsNative classes *)
 	| Some (csup,params) ->
 		let p = f.cf_name_pos in
 		let i = f.cf_name in
 		let check_field f get_super_field is_overload = try
 			(if is_overload && not (has_class_field_flag f CfOverload) then
-				display_error ctx ("Missing overload declaration for field " ^ i) p);
+				display_error ctx.com ("Missing overload declaration for field " ^ i) p);
 			let f_has_override = has_class_field_flag f CfOverride in
 			let t, f2 = get_super_field csup i in
 			check_native_name_override ctx f f2;
@@ -183,34 +185,35 @@ let check_overriding ctx c f =
 			| _ -> ());
 			if has_class_field_flag f2 CfAbstract then begin
 				if f_has_override then
-					display_error ctx ("Field " ^ i ^ " is declared 'override' but parent field " ^ i ^ " is 'abstract' and does not provide any implementation to override") p
+					display_error ctx.com ("Field " ^ i ^ " is declared 'override' but parent field " ^ i ^ " is 'abstract' and does not provide any implementation to override") p
 				else
 					add_class_field_flag f CfOverride (* our spec requires users to not "override" abstract functions, but our implementation depends on implementations to be declared with "override" ¯\_(ツ)_/¯ *)
 			end;
 			if (has_class_field_flag f2 CfOverload && not (has_class_field_flag f CfOverload)) then
-				display_error ctx ("Field " ^ i ^ " should be declared with overload since it was already declared as overload in superclass") p
+				display_error ctx.com ("Field " ^ i ^ " should be declared with overload since it was already declared as overload in superclass") p
 			else if not f_has_override && not (has_class_field_flag f2 CfAbstract) then begin
 				if has_class_flag c CExtern then add_class_field_flag f CfOverride
-				else display_error ctx ("Field " ^ i ^ " should be declared with 'override' since it is inherited from superclass " ^ s_type_path csup.cl_path) p
+				else display_error ctx.com ("Field " ^ i ^ " should be declared with 'override' since it is inherited from superclass " ^ s_type_path csup.cl_path) p
 			end else if not (has_class_field_flag f CfPublic) && (has_class_field_flag f2 CfPublic) then
-				display_error ctx ("Field " ^ i ^ " has less visibility (public/private) than superclass one") p
+				display_error ctx.com ("Field " ^ i ^ " has less visibility (public/private) than superclass one") p
 			else (match f.cf_kind, f2.cf_kind with
 			| _, Method MethInline ->
-				display_error ctx ("Field " ^ i ^ " is inlined and cannot be overridden") p
+				display_error ctx.com ("Field " ^ i ^ " is inlined and cannot be overridden") p
 			| a, b when a = b -> ()
 			| Method MethInline, Method MethNormal ->
 				() (* allow to redefine a method as inlined *)
 			| _ ->
-				display_error ctx ("Field " ^ i ^ " has different property access than in superclass") p);
-			if (has_class_field_flag f2 CfFinal) then display_error ctx ("Cannot override final method " ^ i) p;
+				display_error ctx.com ("Field " ^ i ^ " has different property access than in superclass") p);
+			if (has_class_field_flag f2 CfFinal) then display_error ctx.com ("Cannot override final method " ^ i) p;
 			try
 				let t = apply_params csup.cl_params params t in
-				valid_redefinition ctx f f.cf_type f2 t;
+				let map = TClass.get_map_function csup params in
+				valid_redefinition ctx map map f f.cf_type f2 t;
 			with
 				Unify_error l ->
-					display_error ctx ("Field " ^ i ^ " overrides parent class with different or incomplete type") p;
-					display_error ctx (compl_msg "Base field is defined here") f2.cf_name_pos;
-					display_error ctx (compl_msg (error_msg (Unify l))) p;
+					display_error ctx.com ("Field " ^ i ^ " overrides parent class with different or incomplete type") p;
+					display_error ~depth:1 ctx.com (compl_msg "Base field is defined here") f2.cf_name_pos;
+					located_display_error ~depth:1 ctx.com (compl_located_msg (error_msg p (Unify l)));
 		with
 			Not_found ->
 				if has_class_field_flag f CfOverride then
@@ -224,7 +227,7 @@ let check_overriding ctx c f =
 						) fields [] in
 						StringError.string_error i fields ("Field " ^ i ^ " is declared 'override' but doesn't override any field")
 					end in
-					display_error ctx msg p
+					display_error ctx.com msg p
 		in
 		if has_class_field_flag f CfOverload then begin
 			let overloads = Overloads.get_overloads ctx.com csup i in
@@ -232,7 +235,7 @@ let check_overriding ctx c f =
 				(* check if any super class fields are vars *)
 				match f2.cf_kind with
 				| Var _ ->
-					display_error ctx ("A variable named '" ^ f2.cf_name ^ "' was already declared in a superclass") f.cf_pos
+					display_error ctx.com ("A variable named '" ^ f2.cf_name ^ "' was already declared in a superclass") f.cf_pos
 				| _ -> ()
 			) overloads;
 			List.iter (fun f ->
@@ -251,7 +254,7 @@ let check_overriding ctx c f =
 let class_field_no_interf c i =
 	try
 		let f = PMap.find i c.cl_fields in
-		f.cf_type , f
+		(fun t -> t),f.cf_type , f
 	with Not_found ->
 		match c.cl_super with
 		| None ->
@@ -259,19 +262,14 @@ let class_field_no_interf c i =
 		| Some (c,tl) ->
 			(* rec over class_field *)
 			let _, t , f = raw_class_field (fun f -> f.cf_type) c tl i in
-			apply_params c.cl_params tl t , f
+			let map = TClass.get_map_function c tl in
+			map,apply_params c.cl_params tl t , f
 
 let rec return_flow ctx e =
 	let error() =
-		display_error ctx (Printf.sprintf "Missing return: %s" (s_type (print_context()) ctx.ret)) e.epos; raise Exit
+		display_error ctx.com (Printf.sprintf "Missing return: %s" (s_type (print_context()) ctx.ret)) e.epos; raise Exit
 	in
 	let return_flow = return_flow ctx in
-	let rec uncond e = match e.eexpr with
-		| TIf _ | TWhile _ | TSwitch _ | TTry _ | TFunction _ -> ()
-		| TReturn _ | TThrow _ -> raise Exit
-		| _ -> Type.iter uncond e
-	in
-	let has_unconditional_flow e = try uncond e; false with Exit -> true in
 	match e.eexpr with
 	| TReturn _ | TThrow _ -> ()
 	| TParenthesis e | TMeta(_,e) ->
@@ -280,7 +278,7 @@ let rec return_flow ctx e =
 		let rec loop = function
 			| [] -> error()
 			| [e] -> return_flow e
-			| e :: _ when has_unconditional_flow e -> ()
+			| e :: _ when DeadEnd.has_dead_end e -> ()
 			| _ :: l -> loop l
 		in
 		loop el
@@ -305,7 +303,7 @@ let rec return_flow ctx e =
 		in
 		loop e
 	| _ ->
-		error()
+		if not (DeadEnd.has_dead_end e) then error()
 
 let check_global_metadata ctx meta f_add mpath tpath so =
 	let sl1 = full_dot_path2 mpath tpath in
@@ -319,12 +317,16 @@ let check_global_metadata ctx meta f_add mpath tpath so =
 let check_module_types ctx m p t =
 	let t = t_infos t in
 	try
-		let m2 = Hashtbl.find ctx.g.types_module t.mt_path in
-		if m.m_path <> m2 && String.lowercase (s_type_path m2) = String.lowercase (s_type_path m.m_path) then error ("Module " ^ s_type_path m2 ^ " is loaded with a different case than " ^ s_type_path m.m_path) p;
-		error ("Type name " ^ s_type_path t.mt_path ^ " is redefined from module " ^ s_type_path m2) p
+		let path2 = ctx.com.type_to_module#find t.mt_path in
+		if m.m_path <> path2 && String.lowercase (s_type_path path2) = String.lowercase (s_type_path m.m_path) then typing_error ("Module " ^ s_type_path path2 ^ " is loaded with a different case than " ^ s_type_path m.m_path) p;
+		let m2 = ctx.com.module_lut#find path2 in
+		let hex1 = Digest.to_hex m.m_extra.m_sign in
+		let hex2 = Digest.to_hex m2.m_extra.m_sign in
+		let s = if hex1 = hex2 then hex1 else Printf.sprintf "was %s, is %s" hex2 hex1 in
+		typing_error (Printf.sprintf "Type name %s is redefined from module %s (%s)" (s_type_path t.mt_path)  (s_type_path path2) s) p
 	with
 		Not_found ->
-			Hashtbl.add ctx.g.types_module t.mt_path m.m_path
+			ctx.com.type_to_module#add t.mt_path m.m_path
 
 module Inheritance = struct
 	let is_basic_class_path path = match path with
@@ -333,29 +335,44 @@ module Inheritance = struct
 
 	let check_extends ctx c t p = match follow t with
 		| TInst (csup,params) ->
-			if is_basic_class_path csup.cl_path && not ((has_class_flag c CExtern) && (has_class_flag csup CExtern)) then error "Cannot extend basic class" p;
-			if extends csup c then error "Recursive class" p;
+			if is_basic_class_path csup.cl_path && not ((has_class_flag c CExtern) && (has_class_flag csup CExtern)) then typing_error "Cannot extend basic class" p;
+			if extends csup c then typing_error "Recursive class" p;
 			begin match csup.cl_kind with
 				| KTypeParameter _ ->
-					if is_generic_parameter ctx csup then error "Extending generic type parameters is no longer allowed in Haxe 4" p;
-					error "Cannot extend type parameters" p
+					if is_generic_parameter ctx csup then typing_error "Extending generic type parameters is no longer allowed in Haxe 4" p;
+					typing_error "Cannot extend type parameters" p
 				| _ -> csup,params
 			end
-		| _ -> error "Should extend by using a class" p
+		| _ -> typing_error "Should extend by using a class" p
 
 	let rec check_interface ctx missing c intf params =
 		List.iter (fun (i2,p2) ->
 			check_interface ctx missing c i2 (List.map (apply_params intf.cl_params params) p2)
 		) intf.cl_implements;
 		let p = c.cl_name_pos in
-		let rec check_field i f =
+		let check_field f =
 			let t = (apply_params intf.cl_params params f.cf_type) in
 			let is_overload = ref false in
+			let make_implicit_field () =
+				let cf = {f with cf_overloads = []; cf_type = apply_params intf.cl_params params f.cf_type} in
+				begin try
+					let cf' = PMap.find cf.cf_name c.cl_fields in
+					ctx.com.overload_cache#remove (c.cl_path,f.cf_name);
+					cf'.cf_overloads <- cf :: cf'.cf_overloads
+				with Not_found ->
+					TClass.add_field c cf
+				end;
+				cf
+			in
+			let is_method () = match f.cf_kind with
+				| Method _ -> true
+				| Var _ -> false
+			in
 			try
-				let t2, f2 = class_field_no_interf c i in
+				let map2, t2, f2 = class_field_no_interf c f.cf_name in
 				let t2, f2 =
 					if f2.cf_overloads <> [] || has_class_field_flag f2 CfOverload then
-						let overloads = Overloads.get_overloads ctx.com c i in
+						let overloads = Overloads.get_overloads ctx.com c f.cf_name in
 						is_overload := true;
 						List.find (fun (t1,f1) -> Overloads.same_overload_args t t1 f f1) overloads
 					else
@@ -370,49 +387,48 @@ module Inheritance = struct
 						| MethMacro -> 2
 					in
 					if (has_class_field_flag f CfPublic) && not (has_class_field_flag f2 CfPublic) && not (Meta.has Meta.CompilerGenerated f.cf_meta) then
-						display_error ctx ("Field " ^ i ^ " should be public as requested by " ^ s_type_path intf.cl_path) p
+						display_error ctx.com ("Field " ^ f.cf_name ^ " should be public as requested by " ^ s_type_path intf.cl_path) p
 					else if not (unify_kind f2.cf_kind f.cf_kind) || not (match f.cf_kind, f2.cf_kind with Var _ , Var _ -> true | Method m1, Method m2 -> mkind m1 = mkind m2 | _ -> false) then
-						display_error ctx ("Field " ^ i ^ " has different property access than in " ^ s_type_path intf.cl_path ^ " (" ^ s_kind f2.cf_kind ^ " should be " ^ s_kind f.cf_kind ^ ")") p
+						display_error ctx.com ("Field " ^ f.cf_name ^ " has different property access than in " ^ s_type_path intf.cl_path ^ " (" ^ s_kind f2.cf_kind ^ " should be " ^ s_kind f.cf_kind ^ ")") p
 					else try
-						valid_redefinition ctx f2 t2 f (apply_params intf.cl_params params f.cf_type)
+						let map1 = TClass.get_map_function  intf params in
+						valid_redefinition ctx map1 map2 f2 t2 f (apply_params intf.cl_params params f.cf_type)
 					with
 						Unify_error l ->
 							if not (Meta.has Meta.CsNative c.cl_meta && (has_class_flag c CExtern)) then begin
-								display_error ctx ("Field " ^ i ^ " has different type than in " ^ s_type_path intf.cl_path) p;
-								display_error ctx (compl_msg "Interface field is defined here") f.cf_pos;
-								display_error ctx (compl_msg (error_msg (Unify l))) p;
+								display_error ctx.com ("Field " ^ f.cf_name ^ " has different type than in " ^ s_type_path intf.cl_path) p;
+								display_error ~depth:1 ctx.com (compl_msg "Interface field is defined here") f.cf_pos;
+								located_display_error ~depth:1 ctx.com (compl_located_msg (error_msg p (Unify l)));
 							end
 				)
-			with
-				| Not_found when (has_class_flag c CAbstract) ->
-					let cf = {f with cf_overloads = []} in
+			with Not_found ->
+				if (has_class_flag c CAbstract) && is_method() then begin
+					let cf = make_implicit_field () in
 					add_class_field_flag cf CfAbstract;
-					begin try
-						let cf' = PMap.find cf.cf_name c.cl_fields in
-						Hashtbl.remove ctx.com.overload_cache (c.cl_path,i);
-						cf'.cf_overloads <- cf :: cf'.cf_overloads
-					with Not_found ->
-						TClass.add_field c cf
-					end
-				| Not_found when not (has_class_flag c CInterface) ->
-					if Diagnostics.is_diagnostics_run ctx.com c.cl_pos then
+				end else if has_class_field_flag f CfDefault then begin
+					let cf = make_implicit_field () in
+					cf.cf_expr <- None;
+					add_class_field_flag cf CfExtern;
+					add_class_field_flag cf CfOverride;
+				end else if not (has_class_flag c CInterface) then begin
+					if Diagnostics.error_in_diagnostics_run ctx.com c.cl_pos then
 						DynArray.add missing (f,t)
 					else begin
 						let msg = if !is_overload then
 							let ctx = print_context() in
 							let args = match follow f.cf_type with | TFun(args,_) -> String.concat ", " (List.map (fun (n,o,t) -> (if o then "?" else "") ^ n ^ " : " ^ (s_type ctx t)) args) | _ -> die "" __LOC__ in
-							"No suitable overload for " ^ i ^ "( " ^ args ^ " ), as needed by " ^ s_type_path intf.cl_path ^ " was found"
+							"No suitable overload for " ^ f.cf_name ^ "( " ^ args ^ " ), as needed by " ^ s_type_path intf.cl_path ^ " was found"
 						else
-							("Field " ^ i ^ " needed by " ^ s_type_path intf.cl_path ^ " is missing")
+							("Field " ^ f.cf_name ^ " needed by " ^ s_type_path intf.cl_path ^ " is missing")
 						in
-						display_error ctx msg p
+						display_error ctx.com msg p
 					end
-				| Not_found -> ()
+				end
 		in
-		let check_field i cf =
-			check_field i cf;
+		let check_field _ cf =
+			check_field cf;
 			if has_class_field_flag cf CfOverload then
-				List.iter (check_field i) (List.rev cf.cf_overloads)
+				List.iter check_field (List.rev cf.cf_overloads)
 		in
 		PMap.iter check_field intf.cl_fields
 
@@ -461,7 +477,7 @@ module Inheritance = struct
 		match !missing with
 		| [] ->
 			()
-		| l when Diagnostics.is_diagnostics_run ctx.com c.cl_pos ->
+		| l when Diagnostics.error_in_diagnostics_run ctx.com c.cl_pos ->
 			let diag = {
 				mf_pos = c.cl_name_pos;
 				mf_on = TClassDecl c;
@@ -472,8 +488,8 @@ module Inheritance = struct
 			display.module_diagnostics <- MissingFields diag :: display.module_diagnostics
 		| l ->
 			let singular = match l with [_] -> true | _ -> false in
-			display_error ctx (Printf.sprintf "This class extends abstract class %s but doesn't implement the following method%s" (s_type_path csup.cl_path) (if singular then "" else "s")) c.cl_name_pos;
-			display_error ctx (Printf.sprintf "Implement %s or make %s abstract as well" (if singular then "it" else "them") (s_type_path c.cl_path)) c.cl_name_pos;
+			display_error ctx.com (Printf.sprintf "This class extends abstract class %s but doesn't implement the following method%s" (s_type_path csup.cl_path) (if singular then "" else "s")) c.cl_name_pos;
+			display_error ctx.com (Printf.sprintf "Implement %s or make %s abstract as well" (if singular then "it" else "them") (s_type_path c.cl_path)) c.cl_name_pos;
 			let pctx = print_context() in
 			List.iter (fun (cf,_) ->
 				let s = match follow cf.cf_type with
@@ -482,7 +498,7 @@ module Inheritance = struct
 					| t ->
 						s_type pctx t
 				in
-				display_error ctx (Printf.sprintf "... %s(%s)" cf.cf_name s) cf.cf_name_pos
+				display_error ~depth:1 ctx.com (compl_msg (Printf.sprintf "%s(%s)" cf.cf_name s)) cf.cf_name_pos
 			) (List.rev !missing)
 
 	let set_heritance ctx c herits p =
@@ -496,7 +512,7 @@ module Inheritance = struct
 				| _ -> ()
 			) csup.cl_meta;
 			if has_class_flag csup CFinal && not (((has_class_flag csup CExtern) && Meta.has Meta.Hack c.cl_meta) || (match c.cl_kind with KTypeParameter _ -> true | _ -> false)) then
-				error ("Cannot extend a final " ^ if (has_class_flag c CInterface) then "interface" else "class") p;
+				typing_error ("Cannot extend a final " ^ if (has_class_flag c CInterface) then "interface" else "class") p;
 		in
 		let check_cancel_build csup =
 			match csup.cl_build() with
@@ -521,7 +537,7 @@ module Inheritance = struct
 					let lt = try
 						List.find path_matches ctx.m.curmod.m_types
 					with Not_found ->
-						let t,pi = List.find (fun (lt,_) -> path_matches lt) ctx.m.module_types in
+						let t,pi = List.find (fun (lt,_) -> path_matches lt) ctx.m.module_imports in
 						ImportHandling.mark_import_position ctx pi;
 						t
 					in
@@ -537,18 +553,29 @@ module Inheritance = struct
 		let herits = List.filter (ctx.g.do_inherit ctx c p) herits in
 		(* Pass 1: Check and set relations *)
 		let check_herit t is_extends p =
+			let rec check_interfaces_or_delay () =
+				match c.cl_build() with
+				| BuildMacro pending ->
+					(* Ok listen... we're still building this class, which means we can't check its interfaces yet. However,
+					   we do want to check them at SOME point. So we use this pending list which was maybe designed for this
+					   purpose. However, we STILL have to delay the check because at the time pending is handled, the class
+					   is not built yet. See issue #10847. *)
+					pending := (fun () -> delay ctx PConnectField check_interfaces_or_delay) :: !pending
+				| _ ->
+					check_interfaces ctx c
+			in
 			if is_extends then begin
-				if c.cl_super <> None then error "Cannot extend several classes" p;
+				if c.cl_super <> None then typing_error "Cannot extend several classes" p;
 				let csup,params = check_extends ctx c t p in
 				if (has_class_flag c CInterface) then begin
-					if not (has_class_flag csup CInterface) then error "Cannot extend by using a class" p;
+					if not (has_class_flag csup CInterface) then typing_error "Cannot extend by using a class" p;
 					c.cl_implements <- (csup,params) :: c.cl_implements;
 					if not !has_interf then begin
-						if not is_lib then delay ctx PConnectField (fun() -> check_interfaces ctx c);
+						if not is_lib then delay ctx PConnectField check_interfaces_or_delay;
 						has_interf := true;
 					end
 				end else begin
-					if (has_class_flag csup CInterface) then error "Cannot extend by using an interface" p;
+					if (has_class_flag csup CInterface) then typing_error "Cannot extend by using an interface" p;
 					c.cl_super <- Some (csup,params)
 				end;
 				(fun () ->
@@ -557,16 +584,16 @@ module Inheritance = struct
 				)
 			end else begin match follow t with
 				| TInst ({ cl_path = [],"ArrayAccess" } as ca,[t]) when (has_class_flag ca CExtern) ->
-					if c.cl_array_access <> None then error "Duplicate array access" p;
+					if c.cl_array_access <> None then typing_error "Duplicate array access" p;
 					c.cl_array_access <- Some t;
 					(fun () -> ())
 				| TInst (intf,params) ->
-					if extends intf c then error "Recursive class" p;
-					if (has_class_flag c CInterface) then error "Interfaces cannot implement another interface (use extends instead)" p;
-					if not (has_class_flag intf CInterface) then error "You can only implement an interface" p;
+					if extends intf c then typing_error "Recursive class" p;
+					if (has_class_flag c CInterface) then typing_error "Interfaces cannot implement another interface (use extends instead)" p;
+					if not (has_class_flag intf CInterface) then typing_error "You can only implement an interface" p;
 					c.cl_implements <- (intf, params) :: c.cl_implements;
 					if not !has_interf && not is_lib && not (Meta.has (Meta.Custom "$do_not_check_interf") c.cl_meta) then begin
-						delay ctx PConnectField (fun() -> check_interfaces ctx c);
+						delay ctx PConnectField check_interfaces_or_delay;
 						has_interf := true;
 					end;
 					(fun () ->
@@ -574,19 +601,19 @@ module Inheritance = struct
 						process_meta intf;
 					)
 				| TDynamic t ->
-					if c.cl_dynamic <> None then error "Cannot have several dynamics" p;
-					if not (has_class_flag c CExtern) then display_error ctx "In haxe 4, implements Dynamic is only supported on externs" p;
-					c.cl_dynamic <- Some t;
+					if c.cl_dynamic <> None then typing_error "Cannot have several dynamics" p;
+					if not (has_class_flag c CExtern) then display_error ctx.com "In haxe 4, implements Dynamic is only supported on externs" p;
+					c.cl_dynamic <- Some (match t with None -> t_dynamic | Some t -> t);
 					(fun () -> ())
 				| _ ->
-					error "Should implement by using an interface" p
+					typing_error "Should implement by using an interface" p
 			end
 		in
 		let fl = ExtList.List.filter_map (fun (is_extends,(ct,p)) ->
 			try
 				let t = try
 					Typeload.load_instance ~allow_display:true ctx (ct,p) false
-				with DisplayException(DisplayFields Some({fkind = CRTypeHint} as r)) ->
+				with DisplayException(DisplayFields ({fkind = CRTypeHint} as r)) ->
 					(* We don't allow `implements` on interfaces. Just raise fields completion with no fields. *)
 					if not is_extends && (has_class_flag c CInterface) then raise_fields [] CRImplements r.fsubject;
 					let l = List.filter (fun item -> match item.ci_kind with
@@ -600,8 +627,8 @@ module Inheritance = struct
 					raise_fields l (if is_extends then CRExtends else CRImplements) r.fsubject
 				in
 				Some (check_herit t is_extends p)
-			with Error(Module_not_found(([],name)),p) when ctx.com.display.dms_kind <> DMNone ->
-				if Diagnostics.is_diagnostics_run ctx.com p then DisplayToplevel.handle_unresolved_identifier ctx name p true;
+			with Error(Module_not_found(([],name)),p,_) when ctx.com.display.dms_kind <> DMNone ->
+				if Diagnostics.error_in_diagnostics_run ctx.com p then DisplayToplevel.handle_unresolved_identifier ctx name p true;
 				None
 		) herits in
 		fl
@@ -609,12 +636,22 @@ end
 
 let check_final_vars ctx e =
 	let final_vars = Hashtbl.create 0 in
-	List.iter (fun cf -> match cf.cf_kind with
-		| Var _ when (has_class_field_flag cf CfFinal) && cf.cf_expr = None ->
-			Hashtbl.add final_vars cf.cf_name cf
+	let ordered_fields = DynArray.create () in
+	let rec loop c =
+		List.iter (fun cf -> match cf.cf_kind with
+			| Var _ when (has_class_field_flag cf CfFinal) && cf.cf_expr = None && not (Hashtbl.mem final_vars cf.cf_name) ->
+				Hashtbl.add final_vars cf.cf_name false;
+				DynArray.add ordered_fields (c,cf)
+			| _ ->
+				()
+		) c.cl_ordered_fields;
+		match c.cl_super with
+		| Some(c,_) when has_class_flag c CAbstract && not (has_constructor c) ->
+			loop c
 		| _ ->
 			()
-	) ctx.curclass.cl_ordered_fields;
+	in
+	loop ctx.curclass;
 	if Hashtbl.length final_vars > 0 then begin
 		let rec find_inits e = match e.eexpr with
 			| TBinop(OpAssign,{eexpr = TField({eexpr = TConst TThis},fa)},e2) ->
@@ -624,7 +661,10 @@ let check_final_vars ctx e =
 				Type.iter find_inits e
 		in
 		find_inits e;
-		Hashtbl.iter (fun _ cf ->
-			display_error ctx ("final field " ^ cf.cf_name ^ " must be initialized immediately or in the constructor") cf.cf_pos;
-		) final_vars
-	end
\ No newline at end of file
+		if Hashtbl.length final_vars > 0 then
+			display_error ctx.com "Some final fields are uninitialized in this class" ctx.curclass.cl_name_pos;
+		DynArray.iter (fun (c,cf) ->
+			if Hashtbl.mem final_vars cf.cf_name then
+				display_error ~depth:1 ctx.com "Uninitialized field" cf.cf_name_pos
+		) ordered_fields
+	end
